# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Ramsey XY frequency characterization experiment."""
from typing import List, Optional, Sequence

import numpy as np
from qiskit.circuit import QuantumCircuit, Parameter
from qiskit.providers.backend import Backend
from qiskit.qobj.utils import MeasLevel

from qiskit_experiments.framework import BaseExperiment, Options, BackendTiming
from qiskit_experiments.framework.restless_mixin import RestlessMixin
from qiskit_experiments.library.characterization.analysis import RamseyXYAnalysis


class RamseyXY(BaseExperiment, RestlessMixin):
    r"""A sign-sensitive experiment to measure the frequency of a qubit.

    # section: overview

        This experiment differs from the :class:`~qiskit_experiments.characterization.\
        t2ramsey.T2Ramsey` since it is sensitive to the sign of the frequency offset from
        the main transition. This experiment consists of following two circuits:

        .. parsed-literal::

            (Ramsey X) The second pulse rotates by pi-half around the X axis

                       ┌────┐┌─────────────┐┌───────┐┌────┐ ░ ┌─┐
                  q_0: ┤ √X ├┤ Delay(τ[s]) ├┤ Rz(θ) ├┤ √X ├─░─┤M├
                       └────┘└─────────────┘└───────┘└────┘ ░ └╥┘
            measure: 1/════════════════════════════════════════╩═
                                                               0

            (Ramsey Y) The second pulse rotates by pi-half around the Y axis

                       ┌────┐┌─────────────┐┌───────────┐┌────┐ ░ ┌─┐
                  q_0: ┤ √X ├┤ Delay(τ[s]) ├┤ Rz(θ-π/2) ├┤ √X ├─░─┤M├
                       └────┘└─────────────┘└───────────┘└────┘ ░ └╥┘
            measure: 1/════════════════════════════════════════════╩═
                                                                   0

        The first and second circuits measure the expectation value along the -Y and X axes,
        respectively. This experiment therefore tracks the dynamics of the Bloch vector
        around the equator. The drive frequency of the control electronics defines a reference frame,
        which differs from the true qubit frequency by :math:`\Delta\omega`.
        The Hamiltonian during the
        ``Delay`` instruction is :math:`H^R = - \frac{1}{2} \Delta\omega` in the rotating frame,
        and the propagator will be :math:`U(\tau) = \exp(-iH^R\tau / \hbar)` where :math:`\tau` is the
        duration of the delay. By scanning this duration, we can get

        .. math::

            {\cal E}_x(\tau)
                = {\rm Re} {\rm Tr}\left( Y U \rho U^\dagger \right)
                &= - \cos(\Delta\omega\tau) = \sin(\Delta\omega\tau - \frac{\pi}{2}), \\
            {\cal E}_y(\tau)
                = {\rm Re} {\rm Tr}\left( X U \rho U^\dagger \right)
                &= \sin(\Delta\omega\tau),

        where :math:`\rho` is prepared by the first :math:`\sqrt{\rm X}` gate. Note that phase
        difference of these two outcomes :math:`{\cal E}_x, {\cal E}_y` depends on the sign and
        the magnitude of the frequency offset :math:`\Delta\omega`. By contrast, the measured
        data in the standard Ramsey experiment does not depend on the sign of :math:`\Delta\omega`,
        because :math:`\cos(-\Delta\omega\tau) = \cos(\Delta\omega\tau)`.

        The experiment also allows users to add a small frequency offset to better resolve
        any oscillations. This is implemented by a virtual Z rotation in the circuits. In the
        circuit above it appears as the delay-dependent angle θ(τ).

    # section: analysis_ref
        :class:`RamseyXYAnalysis`

    # section: example
        .. jupyter-execute::
            :hide-code:

            # Temporary workaround for missing support in Qiskit and qiskit-ibm-runtime
            from qiskit_experiments.test.patching import patch_sampler_test_support
            patch_sampler_test_support()

            # backend
            from qiskit_aer import AerSimulator
            from qiskit_ibm_runtime.fake_provider import FakePerth

            backend = AerSimulator.from_backend(FakePerth())

        .. jupyter-execute::

            import numpy as np
            from qiskit_experiments.library.characterization import RamseyXY

            delays = np.linspace(0, 10.e-7, 101)
            exp = RamseyXY((0,), backend=backend, delays=delays, osc_freq=2.0e6)

            exp_data = exp.run().block_for_results()
            display(exp_data.figure(0))
            exp_data.analysis_results(dataframe=True)
    """

    @classmethod
    def _default_experiment_options(cls) -> Options:
        """Default values for the Ramsey XY experiment.

        Experiment Options:
            delays (list): The list of delays that will be scanned in the experiment, in seconds.
            osc_freq (float): A frequency shift in Hz that will be applied by means of
                a virtual Z rotation to increase the frequency of the measured oscillation.
        """
        options = super()._default_experiment_options()
        options.delays = np.linspace(0, 1.0e-6, 51)
        options.osc_freq = 2e6

        return options

    def __init__(
        self,
        physical_qubits: Sequence[int],
        backend: Optional[Backend] = None,
        delays: Optional[List] = None,
        osc_freq: float = 2e6,
    ):
        """Create new experiment.

        Args:
            physical_qubits: List containing the qubit on which to run the
                Ramsey XY experiment.
            backend: Optional, the backend to run the experiment on.
            delays: The delays to scan, in seconds.
            osc_freq: the oscillation frequency induced by the user through a virtual
                Rz rotation. This quantity is given in Hz.
        """
        super().__init__(physical_qubits, analysis=RamseyXYAnalysis(), backend=backend)

        if delays is None:
            delays = self.experiment_options.delays
        self.set_experiment_options(delays=delays, osc_freq=osc_freq)

    def _pre_circuit(self) -> QuantumCircuit:
        """Return a preparation circuit.

        This method can be overridden by subclasses e.g. to run on transitions other
        than the 0 <-> 1 transition.
        """
        return QuantumCircuit(1)

    def circuits(self) -> List[QuantumCircuit]:
        """Create the circuits for the Ramsey XY characterization experiment.

        Returns:
            A list of circuits with a variable delay.
        """
        timing = BackendTiming(self.backend)

        p_delay = Parameter("delay")

        rotation_angle = 2 * np.pi * self.experiment_options.osc_freq * p_delay

        if timing.delay_unit == "dt":
            rotation_angle = rotation_angle * timing.dt

        # Create the X and Y circuits.
        ram_x = self._pre_circuit()
        ram_x.sx(0)
        ram_x.delay(p_delay, 0, timing.delay_unit)
        ram_x.rz(rotation_angle, 0)
        ram_x.sx(0)
        ram_x.measure_active()

        ram_y = self._pre_circuit()
        ram_y.sx(0)
        ram_y.delay(p_delay, 0, timing.delay_unit)
        ram_y.rz(rotation_angle - np.pi / 2, 0)
        ram_y.sx(0)
        ram_y.measure_active()

        circs = []
        for delay in self.experiment_options.delays:
            delay_dt = timing.round_delay(time=delay)
            delay_sec = timing.delay_time(time=delay)

            assigned_x = ram_x.assign_parameters({p_delay: delay_dt}, inplace=False)
            assigned_x.metadata = {
                "series": "X",
                "xval": delay_sec,
            }
            assigned_y = ram_y.assign_parameters({p_delay: delay_dt}, inplace=False)
            assigned_y.metadata = {
                "series": "Y",
                "xval": delay_sec,
            }

            circs.extend([assigned_x, assigned_y])

        return circs

    def _finalize(self):
        # Set initial guess for sinusoidal offset when meas level is 2.
        # This returns probability P1 thus offset=0.5 is obvious.
        # This guarantees reasonable fit especially when data contains only less than half cycle.
        meas_level = self.run_options.get("meas_level", MeasLevel.CLASSIFIED)
        if meas_level == MeasLevel.CLASSIFIED:
            init_guess = self.analysis.options.get("p0", {})
            if "base" not in init_guess:
                init_guess["base"] = 0.5
            self.analysis.set_options(p0=init_guess)

    def _metadata(self):
        metadata = super()._metadata()
        # Store measurement level and meas return if they have been
        # set for the experiment
        for run_opt in ["meas_level", "meas_return"]:
            if hasattr(self.run_options, run_opt):
                metadata[run_opt] = getattr(self.run_options, run_opt)
        return metadata



# This code is part of Qiskit.
#
# (C) Copyright IBM 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""
Quantum process tomography analysis
"""


from typing import List, Dict, Tuple, Union, Optional, Callable, Sequence
import functools
from collections import defaultdict
import numpy as np

from qiskit.result import marginal_distribution

from qiskit_experiments.exceptions import AnalysisError
from qiskit_experiments.library.tomography.basis import MeasurementBasis, PreparationBasis


def tomography_fitter_data(
    data: List[Dict[str, any]],
    outcome_shape: Optional[Union[np.ndarray, int]] = None,
    m_idx_key: str = "m_idx",
    p_idx_key: str = "p_idx",
    clbits_key: str = "clbits",
    clbits: Optional[List[int]] = None,
    cond_clbits_key: str = "cond_clbits",
    cond_clbits: Optional[List[int]] = None,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """Return a tuple of tomography data arrays.

    Args:
        data: tomography experiment data list.
        outcome_shape: Optional, the shape of measurement outcomes for
                       each measurement index. Default is 2.
        m_idx_key: metadata key for measurement basis index (Default: "m_idx").
        p_idx_key: metadata key for preparation basis index (Default: "p_idx").
        clbits_key: metadata clbit key for marginalizing counts. Used if the
            `clbits` kwarg is not provided (Default: "clbits").
        clbits: Optional, clbit indices for marginalizing counts. If provided
            this overrides `clbits_key` in metadata.
        cond_clbits_key: metadata conditional clbit key for conditioning when
            marginalizing counts. Used if the `cond_clbits` kwarg is not
            provided (Default: "cond_clbits").
        cond_clbits: Optional, clbit indices to condition on when when
            marginalizing counts. If provided this overrides `cond_clbits_key`
            in metadata.

    Returns:
        Tuple of data arrays for tomography fitters of form
        (outcome_data, shot_data, measurement data, preparation data).
    """
    meas_size = None
    prep_size = None

    # Construct marginalized tomography count dicts
    outcome_dict = defaultdict(lambda: defaultdict(lambda: 0))

    for datum in data:
        # Get basis data
        metadata = datum["metadata"]
        meas_element = tuple(metadata[m_idx_key]) if m_idx_key in metadata else tuple()
        prep_element = tuple(metadata[p_idx_key]) if p_idx_key in metadata else tuple()
        if meas_size is None:
            meas_size = len(meas_element)
        if prep_size is None:
            prep_size = len(prep_element)
        basis_key = (meas_element, prep_element)

        # Marginalize counts
        counts = datum["counts"]
        num_cond = 0
        count_clbits = []
        if cond_clbits is None and cond_clbits_key is not None:
            cond_clbits = metadata[cond_clbits_key] or []
        if cond_clbits:
            count_clbits += cond_clbits
            num_cond = len(cond_clbits)
        if clbits is None and clbits_key is not None:
            clbits = metadata[clbits_key] or []
        if clbits:
            count_clbits += clbits
        if count_clbits:
            # The input clbits might come in out of order, sort to ensure we
            # don't permute the output during marginalization
            count_clbits = list(sorted(count_clbits))
            counts = marginal_distribution(counts, count_clbits)

        # Accumulate counts
        combined_counts = outcome_dict[basis_key]
        for key, val in counts.items():
            combined_counts[key.replace(" ", "")] += val

    # Format number of outcomes
    outcome_shape = outcome_shape or 2
    if isinstance(outcome_shape, int):
        outcome_shape = meas_size * (outcome_shape,)
    else:
        outcome_shape = tuple(outcome_shape)
    outcome_size = np.prod(outcome_shape, dtype=int)

    # Construct function for converting count outcome dit-strings into
    # integers based on the specified number of outcomes of the measurement
    # bases on each qubit
    if outcome_size == 1:

        def outcome_func(_):
            return 1

    else:
        outcome_func = _int_outcome_function(outcome_shape)

    # Conditional dimension for conditional measurement in source circuit
    if num_cond:
        cond_shape = 2**num_cond
        cond_mask = sum(1 << i for i in range(num_cond))
    else:
        cond_shape = 1
        cond_mask = 0

    # Initialize and fill data arrays
    num_basis = len(outcome_dict)
    measurement_data = np.zeros((num_basis, meas_size), dtype=int)
    preparation_data = np.zeros((num_basis, prep_size), dtype=int)
    shot_data = np.zeros(num_basis, dtype=int)
    outcome_data = np.zeros((cond_shape, num_basis, outcome_size), dtype=int)
    for i, (basis_key, counts) in enumerate(outcome_dict.items()):
        measurement_data[i] = basis_key[0]
        preparation_data[i] = basis_key[1]
        for outcome, freq in counts.items():
            ioutcome = outcome_func(outcome)
            cond_idx = cond_mask & ioutcome
            meas_outcome = ioutcome >> num_cond
            outcome_data[cond_idx][i][meas_outcome] = freq
            shot_data[i] += freq

    return outcome_data, shot_data, measurement_data, preparation_data


@functools.lru_cache(None)
def _int_outcome_function(outcome_shape: Tuple[int, ...]) -> Callable:
    """Generate function for converting string outcomes to ints"""
    # Recursively extract leading bit(dit)
    if len(set(outcome_shape)) == 1:
        # All outcomes are the same shape, so we can use a constant base
        base = outcome_shape[0]
        return lambda outcome: int(outcome, base)

    # General function where each dit could be a different base
    @functools.lru_cache(2048)
    def _int_outcome_general(outcome: str):
        """Convert a general dit-string outcome to integer"""
        # Recursively extract leading bit(dit)
        value = 0
        for i, base in zip(outcome, outcome_shape):
            value *= base
            value += int(i, base)
        return value

    return _int_outcome_general


def _basis_dimensions(
    basis: Optional[Union[MeasurementBasis, PreparationBasis]] = None,
    qubits: Optional[Sequence[int]] = None,
    conditional_indices: Optional[Sequence[int]] = None,
) -> Tuple[int, ...]:
    """Calculate input and output dimensions for basis and qubits"""
    if not qubits:
        return (1,)

    # Get dimension of the preparation and measurement qubits subsystems
    if conditional_indices is None:
        full_qubits = qubits
    else:
        # Remove conditional qubits from full meas qubits
        full_qubits = [q for i, q in enumerate(qubits) if i not in conditional_indices]

    if full_qubits:
        if not basis:
            raise AnalysisError("No tomography basis provided.")
        return basis.matrix_shape(full_qubits)

    return (1,)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Spectroscopy experiment class for resonators."""

import warnings
from typing import Iterable, Optional, Sequence, Tuple
import numpy as np

from qiskit import QuantumCircuit
from qiskit import pulse
from qiskit.circuit import Parameter
from qiskit.exceptions import QiskitError
from qiskit.providers import Backend

from qiskit_experiments.framework import BackendData, BackendTiming, Options
from qiskit_experiments.library.characterization.spectroscopy import Spectroscopy
from qiskit_experiments.database_service import Resonator
from .analysis.resonator_spectroscopy_analysis import ResonatorSpectroscopyAnalysis


class ResonatorSpectroscopy(Spectroscopy):
    """An experiment to perform frequency spectroscopy of the readout resonator.

    # section: overview
        This experiment does spectroscopy on the readout resonator. It applies the following
        circuit

        .. parsed-literal::

                 ┌─┐
              q: ┤M├
                 └╥┘
            c: 1/═╩═
                  0

        where a spectroscopy pulse is attached to the measurement instruction. An initial circuit
        can be added before the measurement by setting the ``initial_circuit`` experiment option. If
        set, the experiment applies the following circuit:

        .. parsed-literal::

                 ┌─────────────────┐┌─┐
              q: ┤ initial_circuit ├┤M├
                 └─────────────────┘└╥┘
            c: 1/════════════════════╩═
                                     0

        Side note 1: when doing readout resonator spectroscopy, each measured IQ point has a
        frequency dependent phase. Close to the resonance, the IQ points start rotating around
        in the IQ plan. This effect must be accounted for in the data processing to produce a
        meaningful signal. The default data processing workflow will therefore reduce the two-
        dimensional IQ data to one-dimensional data using the magnitude of each IQ point.

        Side node 2: when running readout resonator spectroscopy in a parallel experiment the
        user will need to specify the memory slot to use. This can easily be done with the code
        shown below.

        .. code::

            specs = []
            for slot, qubit in enumerate(qubits):
                specs.append(ResonatorSpectroscopy(
                    physical_qubits=[qubit],
                    backend=backend2,
                    memory_slot=slot,
                ))

            exp = ParallelExperiment(specs, backend=backend2)

        # section: warning
            Some backends may not have the required functionality to properly support resonator
            spectroscopy experiments. The experiment may not work or the resulting resonance
            may not properly reflect the properties of the readout resonator.

    # section: example

        The resonator spectroscopy experiment can be run by doing:

        .. code:: python

            qubit = 1
            spec = ResonatorSpectroscopy([qubit], backend)
            exp_data = spec.run().block_for_results()
            exp_data.figure(0)

        This will measure the resonator attached to qubit 1 and report the resonance frequency
        as well as the kappa, i.e. the line width, of the resonator.

    # section: analysis_ref
        :class:`ResonatorSpectroscopyAnalysis`
    """

    @classmethod
    def _default_experiment_options(cls) -> Options:
        """Default option values used for the spectroscopy pulse.

        All units of the resonator spectroscopy experiment are given in seconds.

        Experiment Options:
            amp (float): The amplitude of the spectroscopy pulse. Defaults to 1 and must
                be between 0 and 1.
            duration (float): The duration in seconds of the spectroscopy pulse.
            sigma (float): The standard deviation of the spectroscopy pulse in seconds.
            width (float): The width of the flat-top part of the GaussianSquare pulse in
                seconds.
            initial_circuit (QuantumCircuit): A single-qubit initial circuit to run before the
                measurement/spectroscopy pulse. The circuit must contain only a single qubit and zero
                classical bits. If None, no circuit is appended before the measurement. Defaults to None.
            memory_slot (int): The memory slot that the acquire instruction uses in the pulse schedule.
                The default value is ``0``. This argument allows the experiment to function in a
                :class:`.ParallelExperiment`.
        """
        options = super()._default_experiment_options()

        options.amp = 1
        options.duration = 480e-9
        options.sigma = 60e-9
        options.width = 360e-9
        options.initial_circuit = None
        options.memory_slot = 0

        return options

    def set_experiment_options(self, **fields):
        # Check that the initial circuit is for a single qubit only.
        if "initial_circuit" in fields:
            initial_circuit = fields["initial_circuit"]
            if (
                initial_circuit is not None
                and initial_circuit.num_qubits != 1
                or initial_circuit.num_clbits != 0
            ):
                raise QiskitError(
                    "Initial circuit must be for exactly one qubit and zero classical bits. Got "
                    f"{initial_circuit.num_qubits} qubits and {initial_circuit.num_clbits} "
                    "classical bits instead."
                )
        return super().set_experiment_options(**fields)

    def __init__(
        self,
        physical_qubits: Sequence[int],
        backend: Optional[Backend] = None,
        frequencies: Optional[Iterable[float]] = None,
        absolute: bool = True,
        **experiment_options,
    ):
        """Initialize a resonator spectroscopy experiment.

        A spectroscopy experiment run by setting the frequency of the readout drive.
        The parameters of the GaussianSquare spectroscopy pulse can be specified at run-time
        through the experiment options.

        Args:
            physical_qubits: List containing the resonator on which to run readout
                spectroscopy.
            backend: Optional, the backend to run the experiment on.
            frequencies: The frequencies to scan in the experiment, in Hz. The default values
                range from -20 MHz to 20 MHz in 51 steps. If the ``absolute`` variable is
                set to True then a center frequency obtained from the backend's defaults is
                added to each value of this range.
            absolute: Boolean to specify if the frequencies are absolute or relative to the
                resonator frequency in the backend. The default value is True.
            experiment_options: Key word arguments used to set the experiment options.

        Raises:
            QiskitError: If no frequencies are given and absolute frequencies are desired and
                no backend is given or the backend does not have default measurement frequencies.
        """
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                message="deprecation of Qiskit Pulse",
                module="qiskit_experiments",
                category=DeprecationWarning,
            )
            analysis = ResonatorSpectroscopyAnalysis()

        if frequencies is None:
            frequencies = np.linspace(-20.0e6, 20.0e6, 51)

            if absolute:
                frequencies += self._get_backend_meas_freq(
                    BackendData(backend) if backend is not None else None,
                    physical_qubits[0],
                )

        super().__init__(
            physical_qubits, frequencies, backend, absolute, analysis, **experiment_options
        )

    @staticmethod
    def _get_backend_meas_freq(backend_data: Optional[BackendData], qubit: int) -> float:
        """Get backend meas_freq with error checking"""
        if backend_data is None:
            raise QiskitError(
                "Cannot automatically compute absolute frequencies without a backend."
            )

        if len(backend_data.meas_freqs) < qubit + 1:
            raise QiskitError(
                "Cannot retrieve default measurement frequencies from backend. "
                "Please set frequencies explicitly or set `absolute` to `False`."
            )
        return backend_data.meas_freqs[qubit]

    @property
    def _backend_center_frequency(self) -> float:
        """Returns the center frequency of the experiment.

        Returns:
            The center frequency of the experiment.

        Raises:
            QiskitError: If the experiment does not have a backend set.
        """
        return self._get_backend_meas_freq(self._backend_data, self.physical_qubits[0])

    def _template_circuit(self) -> QuantumCircuit:
        """Return the template quantum circuit."""
        circuit = QuantumCircuit(1, 1)
        if self.experiment_options.initial_circuit is not None:
            circuit.append(self.experiment_options.initial_circuit, [0])
        circuit.measure(0, 0)

        return circuit

    def _schedule(self) -> Tuple[pulse.ScheduleBlock, Parameter]:
        """Create the spectroscopy schedule."""
        timing = BackendTiming(self.backend)

        if timing.dt is None:
            raise QiskitError(f"{self.__class__.__name__} requires a backend with a dt value.")

        duration = timing.round_pulse(time=self.experiment_options.duration)
        sigma = self.experiment_options.sigma / timing.dt
        width = self.experiment_options.width / timing.dt

        qubit = self.physical_qubits[0]

        freq_param = Parameter("frequency")

        with pulse.build(backend=self.backend, name="spectroscopy") as schedule:
            pulse.shift_frequency(freq_param, pulse.MeasureChannel(qubit))
            pulse.play(
                pulse.GaussianSquare(
                    duration=duration,
                    amp=self.experiment_options.amp,
                    sigma=sigma,
                    width=width,
                ),
                pulse.MeasureChannel(qubit),
            )
            pulse.acquire(duration, qubit, pulse.MemorySlot(self.experiment_options.memory_slot))

        return schedule, freq_param

    def _metadata(self):
        """Update metadata with the resonator components."""
        metadata = super()._metadata()
        metadata["device_components"] = list(map(Resonator, self.physical_qubits))
        return metadata

    def circuits(self):
        """Create the circuit for the spectroscopy experiment.

        The circuits are based on a GaussianSquare pulse and a frequency_shift instruction
        encapsulated in a measurement instruction.

        Returns:
            circuits: The circuits that will run the spectroscopy experiment.
        """
        sched, freq_param = self._schedule()

        circs = []
        for freq in self._frequencies:
            freq_shift = freq - self._backend_center_frequency if self._absolute else freq
            freq_shift = np.round(freq_shift, decimals=3)

            sched_ = sched.assign_parameters({freq_param: freq_shift}, inplace=False)

            circuit = self._template_circuit()
            circuit.add_calibration("measure", self.physical_qubits, sched_)
            self._add_metadata(circuit, freq)

            circs.append(circuit)

        return circs



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Spectroscopy experiment class."""

from typing import Tuple

import numpy as np
from qiskit import QuantumCircuit
from qiskit.circuit import Gate, Parameter
from qiskit.exceptions import QiskitError
from qiskit import pulse

from qiskit_experiments.framework import BackendTiming
from qiskit_experiments.library.characterization.spectroscopy import Spectroscopy


class QubitSpectroscopy(Spectroscopy):
    """A spectroscopy experiment to obtain a frequency sweep of the qubit.

    # section: overview
        The circuits produced by spectroscopy, i.e.

        .. parsed-literal::

                       ┌────────────┐ ░ ┌─┐
                  q_0: ┤ Spec(freq) ├─░─┤M├
                       └────────────┘ ░ └╥┘
            measure: 1/══════════════════╩═
                                         0

        have a spectroscopy pulse-schedule embedded in a spectroscopy gate. The
        pulse-schedule consists of a set frequency instruction followed by a GaussianSquare
        pulse. A list of circuits is generated, each with a different frequency "freq".

    # section: analysis_ref
        :class:`~qiskit_experiments.curve_analysis.ResonanceAnalysis`

    # section: example
        .. jupyter-execute::
            :hide-code:

            import warnings

            warnings.filterwarnings(
                "ignore",
                message=".*Due to the deprecation of Qiskit Pulse.*",
                category=DeprecationWarning,
            )

            # backend
            from qiskit_experiments.test.pulse_backend import SingleTransmonTestBackend
            backend = SingleTransmonTestBackend(5.2e9,-.25e9, 1e9, 0.8e9, 1e4, noise=True, seed=199)

        .. jupyter-execute::

            import numpy as np
            from qiskit_experiments.library.characterization import QubitSpectroscopy

            qubit = 0
            freq01_estimate = backend.defaults().qubit_freq_est[qubit]
            frequencies = np.linspace(freq01_estimate-15e6, freq01_estimate+15e6, 51)

            exp = QubitSpectroscopy(physical_qubits = (qubit,),
                                    frequencies = frequencies,
                                    backend = backend,
                                    )
            exp.set_experiment_options(amp=0.005)

            exp_data = exp.run().block_for_results()
            display(exp_data.figure(0))
            exp_data.analysis_results(dataframe=True)
    """

    __spec_gate_name__ = "Spec"

    @property
    def _backend_center_frequency(self) -> float:
        """Returns the center frequency of the experiment.

        Returns:
            The center frequency of the experiment.

        Raises:
            QiskitError: If the experiment does not have a backend set.
        """
        if self.backend is None:
            raise QiskitError("backend not set. Cannot determine the center frequency.")

        return self._backend_data.drive_freqs[self.physical_qubits[0]]

    def _template_circuit(self, freq_param) -> QuantumCircuit:
        """Return the template quantum circuit."""
        circuit = QuantumCircuit(1)
        circuit.append(Gate(name=self.__spec_gate_name__, num_qubits=1, params=[freq_param]), (0,))
        circuit.measure_active()

        return circuit

    def _schedule(self) -> Tuple[pulse.ScheduleBlock, Parameter]:
        """Create the spectroscopy schedule."""
        timing = BackendTiming(self.backend)

        if timing.dt is None:
            raise QiskitError(f"{self.__class__.__name__} requires a backend with a dt value.")

        duration = timing.round_pulse(time=self.experiment_options.duration)
        sigma = self.experiment_options.sigma / timing.dt
        width = self.experiment_options.width / timing.dt

        freq_param = Parameter("frequency")

        with pulse.build(backend=self.backend, name="spectroscopy") as schedule:
            pulse.shift_frequency(freq_param, pulse.DriveChannel(self.physical_qubits[0]))
            pulse.play(
                pulse.GaussianSquare(
                    duration=duration,
                    amp=self.experiment_options.amp,
                    sigma=sigma,
                    width=width,
                ),
                pulse.DriveChannel(self.physical_qubits[0]),
            )
            pulse.shift_frequency(-freq_param, pulse.DriveChannel(self.physical_qubits[0]))

        return schedule, freq_param

    def circuits(self):
        """Create the circuit for the spectroscopy experiment.

        The circuits are based on a GaussianSquare pulse and a frequency_shift instruction
        encapsulated in a gate.

        Returns:
            circuits: The circuits that will run the spectroscopy experiment.
        """

        # Create a template circuit
        sched, freq_param = self._schedule()
        circuit = self._template_circuit(freq_param)
        circuit.add_calibration(
            self.__spec_gate_name__, self.physical_qubits, sched, params=[freq_param]
        )

        # Create the circuits to run
        circs = []
        for freq in self._frequencies:
            freq_shift = freq - self._backend_center_frequency if self._absolute else freq
            freq_shift = np.round(freq_shift, decimals=3)

            assigned_circ = circuit.assign_parameters({freq_param: freq_shift}, inplace=False)
            self._add_metadata(assigned_circ, freq)

            circs.append(assigned_circ)

        return circs



# This code is part of Qiskit.
#
# (C) Copyright IBM 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""Stark Ramsey experiment directly scanning Stark amplitude."""

from __future__ import annotations

import warnings
from collections.abc import Sequence

import numpy as np
from qiskit import pulse
from qiskit.circuit import QuantumCircuit, Gate, ParameterExpression, Parameter
from qiskit.providers.backend import Backend
from qiskit.utils import optionals as _optional
from qiskit.utils.deprecation import deprecate_func

from qiskit_experiments.framework import BaseExperiment, Options, BackendTiming
from .ramsey_amp_scan_analysis import StarkRamseyXYAmpScanAnalysis

if _optional.HAS_SYMENGINE:
    import symengine as sym
else:
    import sympy as sym


class StarkRamseyXYAmpScan(BaseExperiment):
    r"""A fast characterization of Stark frequency shift by varying Stark tone amplitude.

    # section: overview

        This experiment scans Stark tone amplitude at a fixed tone duration.
        The experimental circuits are identical to the :class:`.StarkRamseyXY` experiment
        except that the Stark pulse amplitude is the scanned parameter rather than the pulse width.

        .. parsed-literal::

            (Ramsey X)  The pulse before measurement rotates by pi-half around the X axis

                     ┌────┐┌───────────────────┐┌───┐┌───────────────────┐┌────────┐┌────┐┌─┐
                  q: ┤ √X ├┤ StarkV(stark_amp) ├┤ X ├┤ StarkU(stark_amp) ├┤ Rz(-π) ├┤ √X ├┤M├
                     └────┘└───────────────────┘└───┘└───────────────────┘└────────┘└────┘└╥┘
                c: 1/══════════════════════════════════════════════════════════════════════╩═
                                                                                           0

            (Ramsey Y) The pulse before measurement rotates by pi-half around the Y axis

                     ┌────┐┌───────────────────┐┌───┐┌───────────────────┐┌───────────┐┌────┐┌─┐
                  q: ┤ √X ├┤ StarkV(stark_amp) ├┤ X ├┤ StarkU(stark_amp) ├┤ Rz(-3π/2) ├┤ √X ├┤M├
                     └────┘└───────────────────┘└───┘└───────────────────┘└───────────┘└────┘└╥┘
                c: 1/═════════════════════════════════════════════════════════════════════════╩═
                                                                                              0

        The AC Stark effect can be used to shift the frequency of a qubit with a microwave drive.
        To calibrate a specific frequency shift, the :class:`.StarkRamseyXY` experiment can be run
        to scan the Stark pulse duration at every amplitude, but such a two dimensional scan of
        the tone duration and amplitude may require many circuit executions.
        To avoid this overhead, the :class:`.StarkRamseyXYAmpScan` experiment fixes the
        tone duration and scans only amplitude.

        Recall that an observed Ramsey oscillation in each quadrature may be represented by

        .. math::

            {\cal E}_X(\Omega, t_S) = A e^{-t_S/\tau} \cos \left( 2\pi f_S(\Omega) t_S \right), \\
            {\cal E}_Y(\Omega, t_S) = A e^{-t_S/\tau} \sin \left( 2\pi f_S(\Omega) t_S \right),

        where :math:`f_S(\Omega)` denotes the amount of Stark shift
        at a constant tone amplitude :math:`\Omega`, and :math:`t_S` is the duration of the
        applied tone. For a fixed tone duration,
        one can still observe the Ramsey oscillation by scanning the tone amplitude.
        However, since :math:`f_S` is usually a higher order polynomial of :math:`\Omega`,
        one must manage to fit the y-data for trigonometric functions with
        phase which non-linearly changes with the x-data.
        The :class:`.StarkRamseyXYAmpScan` experiment thus drastically reduces the number of
        circuits to run in return for greater complexity in the fitting model.

    # section: analysis_ref
        :class:`qiskit_experiments.library.driven_freq_tuning.StarkRamseyXYAmpScanAnalysis`

    # section: see_also
        :class:`qiskit_experiments.library.driven_freq_tuning.ramsey.StarkRamseyXY`
        :class:`qiskit_experiments.library.characterization.ramsey_xy.RamseyXY`

    # section: manual
        :doc:`/manuals/characterization/stark_experiment`

    """

    @deprecate_func(
        since="0.8",
        package_name="qiskit-experiments",
        additional_msg=(
            "Due to the deprecation of Qiskit Pulse, experiments involving pulse "
            "gate calibrations like this one have been deprecated."
        ),
    )
    def __init__(
        self,
        physical_qubits: Sequence[int],
        backend: Backend | None = None,
        **experiment_options,
    ):
        """Create new experiment.

        Args:
            physical_qubits: Sequence with the index of the physical qubit.
            backend: Optional, the backend to run the experiment on.
            experiment_options: Experiment options. See the class documentation or
                ``self._default_experiment_options`` for descriptions.
        """
        self._timing = None

        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                message="deprecation of Qiskit Pulse",
                module="qiskit_experiments",
                category=DeprecationWarning,
            )
            analysis = StarkRamseyXYAmpScanAnalysis()

        super().__init__(
            physical_qubits=physical_qubits,
            analysis=analysis,
            backend=backend,
        )
        self.set_experiment_options(**experiment_options)

    @classmethod
    def _default_experiment_options(cls) -> Options:
        """Default experiment options.

        Experiment Options:
            stark_channel (PulseChannel): Pulse channel on which  to apply Stark tones.
                If not provided, the same channel with the qubit drive is used.
                See :ref:`stark_channel_consideration` for details.
            stark_freq_offset (float): Offset of Stark tone frequency from the qubit frequency.
                This offset should be sufficiently large so that the Stark pulse
                does not Rabi drive the qubit.
                See :ref:`stark_frequency_consideration` for details.
            stark_sigma (float): Gaussian sigma of the rising and falling edges
                of the Stark tone, in seconds.
            stark_risefall (float): Ratio of sigma to the duration of
                the rising and falling edges of the Stark tone.
            stark_length (float): Time to accumulate Stark shifted phase in seconds.
            min_stark_amp (float): Minimum Stark tone amplitude.
            max_stark_amp (float): Maximum Stark tone amplitude.
            num_stark_amps (int): Number of Stark tone amplitudes to scan.
            stark_amps (list[float]): The list of amplitude that will be scanned in the experiment.
                If not set, then ``num_stark_amps`` evenly spaced amplitudes
                between ``min_stark_amp`` and ``max_stark_amp`` are used. If ``stark_amps``
                is set, these parameters are ignored.
        """
        options = super()._default_experiment_options()
        options.update_options(
            stark_channel=None,
            stark_freq_offset=80e6,
            stark_sigma=15e-9,
            stark_risefall=2,
            stark_length=50e-9,
            min_stark_amp=-1.0,
            max_stark_amp=1.0,
            num_stark_amps=101,
            stark_amps=None,
        )
        options.set_validator("stark_freq_offset", (0, np.inf))
        options.set_validator("stark_channel", pulse.channels.PulseChannel)
        return options

    def _set_backend(self, backend: Backend):
        super()._set_backend(backend)
        self._timing = BackendTiming(backend)

    def parameters(self) -> np.ndarray:
        """Stark tone amplitudes to use in circuits.

        Returns:
            The list of amplitudes to use for the different circuits based on the
            experiment options.
        """
        opt = self.experiment_options  # alias

        if opt.stark_amps is None:
            params = np.linspace(opt.min_stark_amp, opt.max_stark_amp, opt.num_stark_amps)
        else:
            params = np.asarray(opt.stark_amps, dtype=float)

        return params

    def parameterized_circuits(self) -> tuple[QuantumCircuit, ...]:
        """Create circuits with parameters for Ramsey XY experiment with Stark tone.

        Returns:
            Quantum template circuits for Ramsey X and Ramsey Y experiment.
        """
        opt = self.experiment_options  # alias
        param = Parameter("stark_amp")
        sym_param = param._symbol_expr

        # Pulse gates
        stark_v = Gate("StarkV", 1, [param])
        stark_u = Gate("StarkU", 1, [param])

        # Note that Stark tone yields negative (positive) frequency shift
        # when the Stark tone frequency is higher (lower) than qubit f01 frequency.
        # This choice gives positive frequency shift with positive Stark amplitude.
        qubit_f01 = self._backend_data.drive_freqs[self.physical_qubits[0]]
        neg_sign_of_amp = ParameterExpression(
            symbol_map={param: sym_param},
            expr=-sym.sign(sym_param),
        )
        abs_of_amp = ParameterExpression(
            symbol_map={param: sym_param},
            expr=sym.Abs(sym_param),
        )
        stark_freq = qubit_f01 + neg_sign_of_amp * opt.stark_freq_offset
        stark_channel = opt.stark_channel or pulse.DriveChannel(self.physical_qubits[0])
        ramps_dt = self._timing.round_pulse(time=2 * opt.stark_risefall * opt.stark_sigma)
        sigma_dt = ramps_dt / 2 / opt.stark_risefall
        width_dt = self._timing.round_pulse(time=opt.stark_length)

        with pulse.build() as stark_v_schedule:
            pulse.set_frequency(stark_freq, stark_channel)
            pulse.play(
                pulse.Gaussian(
                    duration=ramps_dt,
                    amp=abs_of_amp,
                    sigma=sigma_dt,
                ),
                stark_channel,
            )

        with pulse.build() as stark_u_schedule:
            pulse.set_frequency(stark_freq, stark_channel)
            pulse.play(
                pulse.GaussianSquare(
                    duration=ramps_dt + width_dt,
                    amp=abs_of_amp,
                    sigma=sigma_dt,
                    risefall_sigma_ratio=opt.stark_risefall,
                ),
                stark_channel,
            )

        ram_x = QuantumCircuit(1, 1)
        ram_x.sx(0)
        ram_x.append(stark_v, [0])
        ram_x.x(0)
        ram_x.append(stark_u, [0])
        ram_x.rz(-np.pi, 0)
        ram_x.sx(0)
        ram_x.measure(0, 0)
        ram_x.metadata = {"series": "X"}
        ram_x.add_calibration(
            gate=stark_v,
            qubits=self.physical_qubits,
            schedule=stark_v_schedule,
        )
        ram_x.add_calibration(
            gate=stark_u,
            qubits=self.physical_qubits,
            schedule=stark_u_schedule,
        )

        ram_y = QuantumCircuit(1, 1)
        ram_y.sx(0)
        ram_y.append(stark_v, [0])
        ram_y.x(0)
        ram_y.append(stark_u, [0])
        ram_y.rz(-np.pi * 3 / 2, 0)
        ram_y.sx(0)
        ram_y.measure(0, 0)
        ram_y.metadata = {"series": "Y"}
        ram_y.add_calibration(
            gate=stark_v,
            qubits=self.physical_qubits,
            schedule=stark_v_schedule,
        )
        ram_y.add_calibration(
            gate=stark_u,
            qubits=self.physical_qubits,
            schedule=stark_u_schedule,
        )

        return ram_x, ram_y

    def circuits(self) -> list[QuantumCircuit]:
        """Create circuits.

        Returns:
            A list of circuits with a variable Stark tone amplitudes.
        """
        ramx_circ, ramy_circ = self.parameterized_circuits()
        param = next(iter(ramx_circ.parameters))

        circs = []
        for amp in self.parameters():
            # Add metadata "direction" to ease the filtering of the data
            # by curve analysis. Indeed, the fit parameters are amplitude sign dependent.

            ramx_circ_assigned = ramx_circ.assign_parameters({param: amp}, inplace=False)
            ramx_circ_assigned.metadata["xval"] = amp
            ramx_circ_assigned.metadata["direction"] = "pos" if amp > 0 else "neg"

            ramy_circ_assigned = ramy_circ.assign_parameters({param: amp}, inplace=False)
            ramy_circ_assigned.metadata["xval"] = amp
            ramy_circ_assigned.metadata["direction"] = "pos" if amp > 0 else "neg"

            circs.extend([ramx_circ_assigned, ramy_circ_assigned])

        return circs

    def _metadata(self) -> dict[str, any]:
        """Return experiment metadata for ExperimentData."""
        metadata = super()._metadata()
        metadata["stark_length"] = self._timing.pulse_time(
            time=self.experiment_options.stark_length
        )
        metadata["stark_freq_offset"] = self.experiment_options.stark_freq_offset

        return metadata



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Base class for calibration-type experiments."""

from abc import ABC, abstractmethod
import functools
import logging
from typing import List, Optional, Sequence, Type, Union
import warnings

from qiskit import QuantumCircuit
from qiskit.providers.options import Options
from qiskit.pulse import ScheduleBlock
from qiskit.transpiler import StagedPassManager, PassManager, Layout, CouplingMap
from qiskit.transpiler.passes import (
    EnlargeWithAncilla,
    FullAncillaAllocation,
    ApplyLayout,
    SetLayout,
)

from qiskit_experiments.calibration_management.calibrations import Calibrations
from qiskit_experiments.calibration_management.update_library import BaseUpdater
from qiskit_experiments.framework.base_analysis import BaseAnalysis
from qiskit_experiments.framework.base_experiment import BaseExperiment
from qiskit_experiments.framework.experiment_data import ExperimentData
from qiskit_experiments.exceptions import CalibrationError

LOG = logging.getLogger(__name__)


class BaseCalibrationExperiment(BaseExperiment, ABC):
    """A mixin class to create calibration experiments.

    This abstract class extends a characterization experiment by turning it into a
    calibration experiment. Such experiments allow schedule management and updating of an
    instance of :class:`.Calibrations`. Furthermore, calibration experiments also specify
    an auto_update variable which, by default, is set to True. If this variable,
    is True then the run method of the experiment will call :meth:`~.ExperimentData.block_for_results`
    and update the calibrations instance once the backend has returned the data.

    This mixin class inherits from the :class:`.BaseExperiment` class since calibration
    experiments by default call :meth:`~.ExperimentData.block_for_results`. This ensures that the next
    calibration experiment cannot proceed before the calibration parameters have been
    updated. Developers that wish to create a calibration experiment must subclass this
    base class and the characterization experiment. Therefore, developers that use this
    mixin class must pay special attention to their class definition. Indeed, the first
    class should be this mixin and the second class should be the characterization
    experiment since the run method from the mixin must be used. For example, the rough
    frequency calibration experiment is defined as

    .. code-block:: python

        RoughFrequencyCal(BaseCalibrationExperiment, QubitSpectroscopy)

    This ensures that the ``run`` method of :class:`.RoughFrequencyCal` will be the
    run method of the :class:`.BaseCalibrationExperiment` class. Furthermore, developers
    must explicitly call the :meth:`__init__` methods of both parent classes.

    Developers should strive to follow the convention that the first two arguments of
    a calibration experiment are the qubit(s) and the :class:`.Calibrations` instance.

    If the experiment uses custom schedules, which is typically the case, then
    developers may chose to use the :meth:`get_schedules` method when creating the
    circuits for the experiment. If :meth:`get_schedules` is used then the developer
    must override at least one of the following methods used by :meth:`get_schedules`
    to set the schedules:

    #. :meth:`_get_schedules_from_options`

    #. :meth:`_get_schedules_from_calibrations`

    #. :meth:`_get_schedules_from_defaults`

    These methods are called by :meth:`get_schedules`.

    The :meth:`update_calibrations` method is responsible for updating the values of the parameters
    stored in the instance of :class:`.Calibrations`. Here, :class:`BaseCalibrationExperiment`
    provides a default update methodology that subclasses can override if a more elaborate behaviour
    is needed. At the minimum the developer must set the variable :code:`_updater` which
    should have an :code:`update` method and can be chosen from the library
    :mod:`qiskit_experiments.calibration_management.update_library`. See also
    :class:`qiskit_experiments.calibration_management.update_library.BaseUpdater`. If no updater
    is specified the experiment will still run but no update of the calibrations will be performed.
    """

    def __init_subclass__(cls, **kwargs):
        """Warn if BaseCalibrationExperiment is not the first parent."""
        for mro_cls in cls.mro():
            if mro_cls is BaseCalibrationExperiment:
                break
            if issubclass(mro_cls, BaseExperiment) and not issubclass(
                mro_cls, BaseCalibrationExperiment
            ):
                warnings.warn(
                    "Calibration experiments must inherit from BaseCalibrationExperiment "
                    f"before a BaseExperiment subclass: {cls}->{mro_cls}."
                )
                break
        super().__init_subclass__(**kwargs)

    # pylint: disable=super-init-not-called
    def __init__(
        self,
        calibrations: Calibrations,
        *args,
        schedule_name: Optional[str] = None,
        cal_parameter_name: Optional[str] = None,
        updater: Optional[Type[BaseUpdater]] = None,
        auto_update: bool = True,
        **kwargs,
    ):
        """Setup the calibration experiment object.

        Args:
            calibrations: The calibrations instance with which to initialize the experiment.
            args: Arguments for the characterization class.
            schedule_name: An optional string which specifies the name of the schedule in
                the calibrations that will be updated.
            cal_parameter_name: An optional string which specifies the name of the parameter in
                the calibrations that will be updated. If None is given then no parameter will
                be updated. Subclasses may assign default values in their init.
            updater: The updater class that updates the Calibrations instance. Different
                calibration experiments will use different updaters.
            auto_update: If set to True (the default) then the calibrations will automatically be
                updated once the experiment has run and :meth:`.block_for_results` will be called.
            kwargs: Keyword arguments for the characterization class.
        """
        super().__init__(*args, **kwargs)
        self._cals = calibrations
        self._sched_name = schedule_name
        self._param_name = cal_parameter_name
        self._updater = updater
        self.auto_update = auto_update

    @property
    def calibrations(self) -> Calibrations:
        """Return the calibrations."""
        return self._cals

    @property
    def analysis(self) -> Union[BaseAnalysis, None]:
        """Return the analysis instance for the experiment.

        .. note::
            Analysis instance set to calibration experiment is implicitly patched to run
            calibration updater to update the parameters in the calibration table.
        """
        return self._analysis

    @analysis.setter
    def analysis(self, analysis: Union[BaseAnalysis, None]) -> None:
        """Set the analysis instance for the experiment"""
        if analysis is None:
            return

        # Create direct alias to the original run method to avoid infinite recursion.
        # .run method is overruled by the wrapped method
        # and thus .run method cannot be called within the wrapper function.
        analysis_run = getattr(analysis, "run")

        @functools.wraps(analysis_run)
        def _wrap_run_analysis(*args, **kwargs):
            experiment_data = analysis_run(*args, **kwargs)
            if self.auto_update:
                experiment_data.add_analysis_callback(self.update_calibrations)
            return experiment_data

        # Monkey patch run method.
        # This calls update_calibrations immediately after standard analysis.
        # This mechanism allows a composite experiment to invoke updater.
        # Note that the composite experiment only takes circuits from individual experiment
        # and the composite analysis calls analysis.run of each experiment.
        # This is only place the updater function can be called from the composite experiment.
        analysis.run = _wrap_run_analysis
        BaseExperiment.analysis.fset(self, analysis)

    @classmethod
    def _default_experiment_options(cls) -> Options:
        """Default values for a calibration experiment.

        Experiment Options:
            result_index (int): The index of the result from which to update the calibrations.
            group (str): The calibration group to which the parameter belongs. This will default
                to the value "default".
        """
        options = super()._default_experiment_options()
        options.update_options(result_index=-1, group="default")
        return options

    @classmethod
    def _default_transpile_options(cls) -> Options:
        """Return empty default transpile options as optimization_level is not used."""
        return Options()

    def set_transpile_options(self, **fields):
        r"""Add a warning message.

        .. note::
            If your experiment has overridden `_transpiled_circuits` and needs
            transpile options then please also override `set_transpile_options`.
        """
        warnings.warn(f"Transpile options are not used in {self.__class__.__name__ }.")

    def update_calibrations(self, experiment_data: ExperimentData):
        """Update parameter values in the :class:`.Calibrations` instance.

        The default behaviour is to call the update method of the class variable
        :code:`__updater__` with simplistic options. Subclasses can override this
        method to update the instance of :class:`.Calibrations` if they require a
        more sophisticated behaviour as is the case for the :class:`.Rabi` and
        :class:`.FineAmplitude` calibration experiments.
        """
        if self._updater is not None:
            self._updater.update(
                self._cals,
                experiment_data,
                parameter=self._param_name,
                schedule=self._sched_name,
            )

    def _validate_channels(self, schedule: ScheduleBlock, physical_qubits: Sequence[int]):
        """Check that the physical qubits are contained in the schedule.

        This is a helper method that experiment developers can call in their implementation
        of :meth:`validate_schedules` when checking the schedules.

        Args:
            schedule: The schedule for which to check the qubits.
            physical_qubits: The qubits that should be included in the schedule.

        Raises:
            CalibrationError: If a physical qubit is not contained in the channels schedule.
        """
        for qubit in physical_qubits:
            if qubit not in set(ch.index for ch in schedule.channels):
                raise CalibrationError(
                    f"Schedule {schedule.name} does not contain a channel "
                    f"for the physical qubit {qubit}."
                )

    def _validate_parameters(self, schedule: ScheduleBlock, n_expected_parameters: int):
        """Check that the schedule has the expected number of parameters.

        This is a helper method that experiment developers can call in their implementation
        of :meth:`validate_schedules` when checking the schedules.

        Args:
            schedule: The schedule for which to check the qubits.
            n_expected_parameters: The number of free parameters the schedule must have.

        Raises:
            CalibrationError: If the schedule does not have n_expected_parameters parameters.
        """
        if len(schedule.parameters) != n_expected_parameters:
            raise CalibrationError(
                f"The schedules {schedule.name} for {self.__class__.__name__} must have "
                f"{n_expected_parameters} parameters. Found {len(schedule.parameters)}."
            )

    def _metadata(self):
        """Add standard calibration metadata."""
        metadata = super()._metadata()

        metadata["cal_group"] = self.experiment_options.group
        metadata["cal_param_name"] = self._param_name
        metadata["cal_schedule"] = self._sched_name

        # Store measurement level and meas return if they have been
        # set for the experiment
        for run_opt in ["meas_level", "meas_return"]:
            if hasattr(self.run_options, run_opt):
                metadata[run_opt] = getattr(self.run_options, run_opt)
        return metadata

    def _transpiled_circuits(self) -> List[QuantumCircuit]:
        """Override the transpiled circuits method to bring in the inst_map.

        The transpilation should do the strict minimum to make the circuits hardware compatible.
        Indeed, calibration experiments are designed with a specific gate sequence in mind. Any
        transpiler operation that changes this gate sequence may compromise the validity of the
        calibration experiment. Sub-classes may override this method to define their own
        transpilation if need be.

        Returns:
            A list of transpiled circuits.
        """
        transpiled = []
        for circ in self.circuits():
            circ = self._map_to_physical_qubits(circ)
            self._attach_calibrations(circ)

            transpiled.append(circ)

        return transpiled

    def _map_to_physical_qubits(self, circuit: QuantumCircuit) -> QuantumCircuit:
        """Map program qubits to physical qubits.

        Args:
            circuit: The quantum circuit to map to device qubits.

        Returns:
            A quantum circuit that has the same number of qubits as the backend and where
            the physical qubits of the experiment have been properly mapped.
        """
        initial_layout = Layout.from_intlist(list(self.physical_qubits), *circuit.qregs)

        coupling_map = self._backend_data.coupling_map
        if coupling_map is not None:
            coupling_map = CouplingMap(self._backend_data.coupling_map)

        layout = PassManager(
            [
                SetLayout(initial_layout),
                FullAncillaAllocation(coupling_map),
                EnlargeWithAncilla(),
                ApplyLayout(),
            ]
        )

        return StagedPassManager(["layout"], layout=layout).run(circuit)

    @abstractmethod
    def _attach_calibrations(self, circuit: QuantumCircuit):
        """Attach the calibrations to the quantum circuit.

        This method attaches calibrations from the `self._cals` instance to the transpiled
        quantum circuits. Given how important this method is it is made abstract to force
        potential calibration experiment developers to implement it and think about how
        schedules are attached to the circuits. The implementation of this method is delegated
        to the sub-classes so that they can map gate instructions to the schedules stored in the
        ``Calibrations`` instance. This method is needed for most calibration experiments. However,
        some experiments already attach circuits to the logical circuits and do not needed to run
        ``_attach_calibrations``. In such experiments a simple ``pass`` statement will suffice.
        """



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""A qubit to control channel map."""

from typing import Any, Dict, List, Tuple

from qiskit.pulse import ControlChannel


class ControlChannelMap:
    """A class to help serialize control channel maps."""

    def __init__(self, control_channel_map: Dict[Tuple[int, ...], List[ControlChannel]]):
        """Setup the control channel map.

        Args:
            control_channel_map: A configuration dictionary of any control channels. The
                keys are tuples of qubits and the values are a list of ControlChannels
                that correspond to the qubits in the keys.
        """
        self._map = control_channel_map or {}

    @property
    def chan_map(self):
        """Return the qubits to control channel map."""
        return self._map

    def config(self) -> Dict[str, Any]:
        """Return the settings used to initialize the mapping."""
        return {
            "class": self.__class__.__name__,
            "map": [{"key": k, "value": [chan.index for chan in v]} for k, v in self._map.items()],
        }

    @classmethod
    def from_config(cls, config: Dict) -> "ControlChannelMap":
        """Deserialize the control channel map given the input dictionary"""

        ch_map = config["map"]

        return cls(
            {tuple(item["key"]): [ControlChannel(idx) for idx in item["value"]] for item in ch_map}
        )

    def __json_encode__(self):
        """Convert to format that can be JSON serialized."""
        return self.config()

    @classmethod
    def __json_decode__(cls, value: Dict[str, Any]) -> "ControlChannelMap":
        """Load from JSON compatible format."""
        return cls.from_config(value)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""Stark Ramsey experiment."""

from __future__ import annotations

import warnings
from collections.abc import Sequence

import numpy as np
from qiskit import pulse
from qiskit.circuit import QuantumCircuit, Gate, Parameter
from qiskit.providers.backend import Backend
from qiskit.utils import optionals as _optional
from qiskit.utils.deprecation import deprecate_func

from qiskit_experiments.framework import BaseExperiment, Options, BackendTiming
from qiskit_experiments.library.characterization.analysis import RamseyXYAnalysis

if _optional.HAS_SYMENGINE:
    pass
else:
    pass


class StarkRamseyXY(BaseExperiment):
    """A sign-sensitive experiment to measure the frequency of a qubit under a pulsed Stark tone.

    # section: overview

        This experiment is a variant of :class:`.RamseyXY` with a pulsed Stark tone
        and consists of the following two circuits:

        .. parsed-literal::

            (Ramsey X)  The pulse before measurement rotates by pi-half around the X axis

                     ┌────┐┌────────┐┌───┐┌───────────────┐┌────────┐┌────┐┌─┐
                  q: ┤ √X ├┤ StarkV ├┤ X ├┤ StarkU(delay) ├┤ Rz(-π) ├┤ √X ├┤M├
                     └────┘└────────┘└───┘└───────────────┘└────────┘└────┘└╥┘
                c: 1/═══════════════════════════════════════════════════════╩═
                                                                            0

            (Ramsey Y) The pulse before measurement rotates by pi-half around the Y axis

                     ┌────┐┌────────┐┌───┐┌───────────────┐┌───────────┐┌────┐┌─┐
                  q: ┤ √X ├┤ StarkV ├┤ X ├┤ StarkU(delay) ├┤ Rz(-3π/2) ├┤ √X ├┤M├
                     └────┘└────────┘└───┘└───────────────┘└───────────┘└────┘└╥┘
                c: 1/══════════════════════════════════════════════════════════╩═
                                                                               0

        In principle, the sequence is a variant of :class:`.RamseyXY` circuit.
        However, the delay in between √X gates is replaced with an off-resonant drive.
        This off-resonant drive shifts the qubit frequency due to the
        Stark effect and causes it to accumulate phase during the
        Ramsey sequence. This frequency shift is a function of the
        offset of the Stark tone frequency from the qubit frequency
        and of the magnitude of the tone.

        Note that the Stark tone pulse (StarkU) takes the form of a flat-topped Gaussian envelope.
        The magnitude of the pulse varies in time during its rising and falling edges.
        It is difficult to characterize the net phase accumulation of the qubit during the
        edges of the pulse when the frequency shift is varying with the pulse amplitude.
        In order to simplify the analysis, an additional pulse (StarkV)
        involving only the edges of StarkU is added to the sequence.
        The sign of the phase accumulation is inverted for StarkV relative to that of StarkU
        by inserting an X gate in between the two pulses.

        This technique allows the experiment to accumulate only the net phase
        during the flat-top part of the StarkU pulse with constant magnitude.

    # section: analysis_ref
        :class:`qiskit_experiments.library.characterization.RamseyXYAnalysis`

    # section: see_also
        :class:`qiskit_experiments.library.characterization.ramsey_xy.RamseyXY`

    # section: manual
        :doc:`/manuals/characterization/stark_experiment`

    """

    @deprecate_func(
        since="0.8",
        package_name="qiskit-experiments",
        additional_msg=(
            "Due to the deprecation of Qiskit Pulse, experiments involving pulse "
            "gate calibrations like this one have been deprecated."
        ),
    )
    def __init__(
        self,
        physical_qubits: Sequence[int],
        backend: Backend | None = None,
        **experiment_options,
    ):
        """Create new experiment.

        Args:
            physical_qubits: Index of physical qubit.
            backend: Optional, the backend to run the experiment on.
            experiment_options: Experiment options. See the class documentation or
                ``self._default_experiment_options`` for descriptions.
        """
        self._timing = None

        super().__init__(
            physical_qubits=physical_qubits,
            analysis=RamseyXYAnalysis(),
            backend=backend,
        )
        self.set_experiment_options(**experiment_options)

    @classmethod
    def _default_experiment_options(cls) -> Options:
        """Default experiment options.

        Experiment Options:
            stark_amp (float): A single float parameter to represent the magnitude of Stark tone
                and the sign of expected Stark shift.
                See :ref:`stark_tone_implementation` for details.
            stark_channel (PulseChannel): Pulse channel to apply Stark tones.
                If not provided, the same channel with the qubit drive is used.
                See :ref:`stark_channel_consideration` for details.
            stark_freq_offset (float): Offset of Stark tone frequency from the qubit frequency.
                This offset should be sufficiently large so that the Stark pulse
                does not Rabi drive the qubit.
                See :ref:`stark_frequency_consideration` for details.
            stark_sigma (float): Gaussian sigma of the rising and falling edges
                of the Stark tone, in seconds.
            stark_risefall (float): Ratio of sigma to the duration of
                the rising and falling edges of the Stark tone.
            min_freq (float): Minimum frequency that this experiment is guaranteed to resolve.
                Note that fitter algorithm :class:`.RamseyXYAnalysis` of this experiment
                is still capable of fitting experiment data with lower frequency.
            max_freq (float): Maximum frequency that this experiment can resolve.
            delays (list[float]): The list of delays if set that will be scanned in the
                experiment. If not set, then evenly spaced delays with interval
                computed from ``min_freq`` and ``max_freq`` are used.
                See :meth:`StarkRamseyXY.delays` for details.
        """
        options = super()._default_experiment_options()
        options.update_options(
            stark_amp=0.0,
            stark_channel=None,
            stark_freq_offset=80e6,
            stark_sigma=15e-9,
            stark_risefall=2,
            min_freq=5e6,
            max_freq=100e6,
            delays=None,
        )
        options.set_validator("stark_freq_offset", (0, np.inf))
        options.set_validator("stark_channel", pulse.channels.PulseChannel)
        return options

    def _set_backend(self, backend: Backend):
        super()._set_backend(backend)
        self._timing = BackendTiming(backend)

    def set_experiment_options(self, **fields):
        _warning_circuit_length = 300

        # Do validation for circuit number
        min_freq = fields.get("min_freq", None)
        max_freq = fields.get("max_freq", None)
        delays = fields.get("delays", None)
        if min_freq is not None and max_freq is not None:
            if delays:
                warnings.warn(
                    "Experiment option 'min_freq' and 'max_freq' are ignored "
                    "when 'delays' are explicitly specified.",
                    UserWarning,
                )
            else:
                n_expr_circs = 2 * int(2 * max_freq / min_freq)  # delays * (x, y)
                max_circs_per_job = None
                if self._backend_data:
                    max_circs_per_job = self._backend_data.max_circuits()
                if n_expr_circs > (max_circs_per_job or _warning_circuit_length):
                    warnings.warn(
                        f"Provided configuration generates {n_expr_circs} circuits. "
                        "You can set smaller 'max_freq' or larger 'min_freq' to reduce this number. "
                        "This experiment is still executable but your execution may consume "
                        "unnecessary long quantum device time, and result may suffer "
                        "device parameter drift in consequence of the long execution time.",
                        UserWarning,
                    )
        # Do validation for spectrum overlap to avoid real excitation
        stark_freq_offset = fields.get("stark_freq_offset", None)
        stark_sigma = fields.get("stark_sigma", None)
        if stark_freq_offset is not None and stark_sigma is not None:
            if stark_freq_offset < 1 / stark_sigma:
                warnings.warn(
                    "Provided configuration may induce coherent state exchange between qubit levels "
                    "because of the potential spectrum overlap. You can avoid this by "
                    "increasing the 'stark_sigma' or 'stark_freq_offset'. "
                    "Note that this experiment is still executable.",
                    UserWarning,
                )
            pass

        super().set_experiment_options(**fields)

    def parameters(self) -> np.ndarray:
        """Delay values to use in circuits.

        .. note::

            The delays are computed with the ``min_freq`` and ``max_freq`` experiment options.
            The maximum point is computed from the ``min_freq`` to guarantee the result
            contains at least one Ramsey oscillation cycle at this frequency.
            The interval is computed from the ``max_freq`` to sample with resolution
            such that the Nyquist frequency is twice ``max_freq``.

        Returns:
            The list of delays to use for the different circuits based on the
            experiment options.

        Raises:
            ValueError: When ``min_freq`` is larger than ``max_freq``.
        """
        opt = self.experiment_options  # alias

        if opt.delays is None:
            if opt.min_freq > opt.max_freq:
                raise ValueError("Experiment option 'min_freq' must be smaller than 'max_freq'.")
            # Delay is longer enough to capture 1 cycle of the minimum frequency.
            # Fitter can still accurately fit samples shorter than 1 cycle.
            max_period = 1 / opt.min_freq
            # Inverse of interval should be greater than Nyquist frequency.
            sampling_freq = 2 * opt.max_freq
            interval = 1 / sampling_freq
            return np.arange(0, max_period, interval)
        return opt.delays

    def parameterized_circuits(self) -> tuple[QuantumCircuit, ...]:
        """Create circuits with parameters for Ramsey XY experiment with Stark tone.

        Returns:
            Quantum template circuits for Ramsey X and Ramsey Y experiment.
        """
        opt = self.experiment_options  # alias
        param = Parameter("delay")

        # Pulse gates
        stark_v = Gate("StarkV", 1, [])
        stark_u = Gate("StarkU", 1, [param])

        # Note that Stark tone yields negative (positive) frequency shift
        # when the Stark tone frequency is higher (lower) than qubit f01 frequency.
        # This choice gives positive frequency shift with positive Stark amplitude.
        qubit_f01 = self._backend_data.drive_freqs[self.physical_qubits[0]]
        stark_freq = qubit_f01 - np.sign(opt.stark_amp) * opt.stark_freq_offset
        stark_amp = np.abs(opt.stark_amp)
        stark_channel = opt.stark_channel or pulse.DriveChannel(self.physical_qubits[0])
        ramps_dt = self._timing.round_pulse(time=2 * opt.stark_risefall * opt.stark_sigma)
        sigma_dt = ramps_dt / 2 / opt.stark_risefall

        with pulse.build() as stark_v_schedule:
            pulse.set_frequency(stark_freq, stark_channel)
            pulse.play(
                pulse.Gaussian(
                    duration=ramps_dt,
                    amp=stark_amp,
                    sigma=sigma_dt,
                    name="StarkV",
                ),
                stark_channel,
            )

        with pulse.build() as stark_u_schedule:
            pulse.set_frequency(stark_freq, stark_channel)
            pulse.play(
                pulse.GaussianSquare(
                    duration=ramps_dt + param,
                    amp=stark_amp,
                    sigma=sigma_dt,
                    risefall_sigma_ratio=opt.stark_risefall,
                    name="StarkU",
                ),
                stark_channel,
            )

        ram_x = QuantumCircuit(1, 1)
        ram_x.sx(0)
        ram_x.append(stark_v, [0])
        ram_x.x(0)
        ram_x.append(stark_u, [0])
        ram_x.rz(-np.pi, 0)
        ram_x.sx(0)
        ram_x.measure(0, 0)
        ram_x.metadata = {"series": "X"}
        ram_x.add_calibration(
            gate=stark_v,
            qubits=self.physical_qubits,
            schedule=stark_v_schedule,
        )
        ram_x.add_calibration(
            gate=stark_u,
            qubits=self.physical_qubits,
            schedule=stark_u_schedule,
        )

        ram_y = QuantumCircuit(1, 1)
        ram_y.sx(0)
        ram_y.append(stark_v, [0])
        ram_y.x(0)
        ram_y.append(stark_u, [0])
        ram_y.rz(-np.pi * 3 / 2, 0)
        ram_y.sx(0)
        ram_y.measure(0, 0)
        ram_y.metadata = {"series": "Y"}
        ram_y.add_calibration(
            gate=stark_v,
            qubits=self.physical_qubits,
            schedule=stark_v_schedule,
        )
        ram_y.add_calibration(
            gate=stark_u,
            qubits=self.physical_qubits,
            schedule=stark_u_schedule,
        )

        return ram_x, ram_y

    def circuits(self) -> list[QuantumCircuit]:
        """Create circuits.

        Returns:
            A list of circuits with a variable delay.
        """
        timing = BackendTiming(self.backend, min_length=0)

        ramx_circ, ramy_circ = self.parameterized_circuits()
        param = next(iter(ramx_circ.parameters))

        circs = []
        for delay in self.parameters():
            valid_delay_dt = timing.round_pulse(time=delay)
            net_delay_sec = timing.pulse_time(time=delay)

            ramx_circ_assigned = ramx_circ.assign_parameters({param: valid_delay_dt}, inplace=False)
            ramx_circ_assigned.metadata["xval"] = net_delay_sec

            ramy_circ_assigned = ramy_circ.assign_parameters({param: valid_delay_dt}, inplace=False)
            ramy_circ_assigned.metadata["xval"] = net_delay_sec

            circs.extend([ramx_circ_assigned, ramy_circ_assigned])

        return circs

    def _metadata(self) -> dict[str, any]:
        """Return experiment metadata for ExperimentData."""
        metadata = super()._metadata()
        metadata["stark_amp"] = self.experiment_options.stark_amp
        metadata["stark_freq_offset"] = self.experiment_options.stark_freq_offset

        return metadata



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021, 2022, 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
r"""
=========================================================
Visualization (:mod:`qiskit_experiments.visualization`)
=========================================================

.. currentmodule:: qiskit_experiments.visualization

The visualization module provides plotting functionality for creating figures from
experiment and analysis results. This includes plotter and drawer classes to plot data
in :class:`.CurveAnalysis` and its subclasses. Plotters inherit from
:class:`BasePlotter` and define a type of figure that may be generated from experiment
or analysis data. For example, the results from :class:`.CurveAnalysis`---or any other
experiment where results are plotted against a single parameter (i.e., :math:`x`)---can
be plotted using the :class:`CurvePlotter` class, which plots X-Y-like values.

These plotter classes act as a bridge (from the common bridge pattern in software
development) between analysis classes (or even users) and plotting backends such as
Matplotlib. Drawers are the backends, with a common interface defined in
:class:`BaseDrawer`. Though Matplotlib is the only officially supported plotting backend
in Qiskit Experiments (i.e., through :class:`MplDrawer`), custom drawers can be
implemented by users to use alternative backends. As long as the backend is a subclass
of :class:`BaseDrawer`, and implements all the necessary functionality, all plotters
should be able to generate figures with the alternative backend.

To collate style parameters together, plotters and drawers store instances of the
:class:`PlotStyle` class. These instances can be merged and updated, so that default
styles can have their values overwritten.

Plotter Library
===============

.. autosummary::
    :toctree: ../stubs/
    :template: autosummary/plotter.rst

    BasePlotter
    CurvePlotter
    IQPlotter

Drawer Library
==============

.. autosummary::
    :toctree: ../stubs/
    :template: autosummary/drawer.rst

    BaseDrawer
    MplDrawer

Plotting Style
==============

.. autosummary::
    :toctree: ../stubs/
    :template: autosummary/class.rst

    PlotStyle
"""

from .drawers import BaseDrawer, MplDrawer
from .plotters import BasePlotter, CurvePlotter, IQPlotter
from .style import PlotStyle



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
A library of parameter guess functions.
"""
# pylint: disable=invalid-name

import functools
from typing import Optional, Tuple, Callable

import numpy as np
from scipy import signal

from qiskit_experiments.exceptions import AnalysisError


def frequency(
    x: np.ndarray,
    y: np.ndarray,
    filter_window: int = 5,
    filter_dim: int = 2,
) -> float:
    r"""Get frequency of oscillating signal.

    First this tries FFT. If the true value is likely below or near the frequency resolution,
    the function tries low frequency fit with

    .. math::

        f_{\rm est} = \frac{1}{2\pi {\rm max}\left| y \right|}
            {\rm max} \left| \frac{dy}{dx} \right|

    given :math:`y = A \cos (2\pi f x + phi)`. In this mode, y data points are
    smoothed by a Savitzky-Golay filter to protect against outlier points.

    .. note::

        This function returns always positive frequency.
        This function is sensitive to the DC offset.
        This function assumes sorted, no-overlapping x values.

    Args:
        x: Array of x values.
        y: Array of y values.
        filter_window: Window size of Savitzky-Golay filter. This should be odd number.
        filter_dim: Dimension of Savitzky-Golay filter.

    Returns:
        Frequency estimation of oscillation signal.
    """
    # to run FFT x interval should be identical
    sampling_interval = np.unique(np.round(np.diff(x), decimals=20))

    if len(sampling_interval) != 1:
        # resampling with minimum xdata interval
        sampling_interval = np.min(sampling_interval)
        x_ = np.arange(x[0], x[-1], sampling_interval)
        y_ = np.interp(x_, xp=x, fp=y)
    else:
        sampling_interval = sampling_interval[0]
        x_ = x
        y_ = y

    fft_data = np.fft.fft(y_ - np.average(y_))
    freqs = np.fft.fftfreq(len(x_), sampling_interval)

    positive_freqs = freqs[freqs >= 0]
    positive_fft_data = fft_data[freqs >= 0]

    freq_guess = positive_freqs[np.argmax(np.abs(positive_fft_data))]

    if freq_guess < 1.5 / (sampling_interval * len(x_)):
        # low frequency fit, use this mode when the estimate is near the resolution
        y_smooth = signal.savgol_filter(y_, window_length=filter_window, polyorder=filter_dim)

        # no offset is assumed
        y_amp = max(np.abs(y_smooth))

        if np.isclose(y_amp, 0.0):
            # no oscillation signal
            return 0.0

        freq_guess = max(np.abs(np.diff(y_smooth) / sampling_interval)) / (y_amp * 2 * np.pi)

    return freq_guess


def max_height(
    y: np.ndarray,
    percentile: Optional[float] = None,
    absolute: bool = False,
) -> Tuple[float, int]:
    """Get maximum value of y curve and its index.

    Args:
        y: Array of y values.
        percentile: Return that percentile value if provided, otherwise just return max value.
        absolute: Use absolute y value.

    Returns:
        The maximum y value and index.
    """
    if percentile is not None:
        return get_height(y, functools.partial(np.percentile, q=percentile), absolute)
    return get_height(y, np.nanmax, absolute)


def min_height(
    y: np.ndarray,
    percentile: Optional[float] = None,
    absolute: bool = False,
) -> Tuple[float, int]:
    """Get minimum value of y curve and its index.

    Args:
        y: Array of y values.
        percentile: Return that percentile value if provided, otherwise just return min value.
        absolute: Use absolute y value.

    Returns:
        The minimum y value and index.
    """
    if percentile is not None:
        return get_height(y, functools.partial(np.percentile, q=percentile), absolute)
    return get_height(y, np.nanmin, absolute)


def get_height(
    y: np.ndarray,
    find_height: Callable,
    absolute: bool = False,
) -> Tuple[float, int]:
    """Get specific value of y curve defined by a callback and its index.

    Args:
        y: Array of y values.
        find_height: A callback to find preferred y value.
        absolute: Use absolute y value.

    Returns:
        The target y value and index.
    """
    if absolute:
        y_ = np.abs(y)
    else:
        y_ = y

    y_target = find_height(y_)
    index = int(np.argmin(np.abs(y_ - y_target)))

    return y_target, index


def exp_decay(x: np.ndarray, y: np.ndarray) -> float:
    r"""Get exponential decay parameter from monotonically increasing (decreasing) curve.

    This assumes following function form.

    .. math::

        y(x) = e^{\alpha x}

    We can calculate :math:`\alpha` as

    .. math::

        \alpha = \log(y(x)) / x

    To find this number, the numpy polynomial fit with ``deg=1`` is used.

    Args:
        x: Array of x values.
        y: Array of y values.

    Returns:
         Decay rate of signal.
    """
    inds = y > 0
    if np.count_nonzero(inds) < 2:
        return 0

    coeffs = np.polyfit(x[inds], np.log(y[inds]), deg=1)

    return float(coeffs[0])


def oscillation_exp_decay(
    x: np.ndarray,
    y: np.ndarray,
    filter_window: int = 5,
    filter_dim: int = 2,
    freq_guess: Optional[float] = None,
) -> float:
    r"""Get exponential decay parameter from oscillating signal.

    This assumes following function form.

    .. math::

        y(x) = e^{\alpha x} F(x),

    where :math:`F(x)` is arbitrary oscillation function oscillating at ``freq_guess``.
    This function first applies a Savitzky-Golay filter to y value,
    then run scipy peak search to extract peak positions.
    If ``freq_guess`` is provided, the search function will be robust to fake peaks due to noise.
    This function calls :func:`exp_decay` function for extracted x and y values at peaks.

    .. note::

        y values should contain more than one cycle of oscillation to use this guess approach.

    Args:
        x: Array of x values.
        y: Array of y values.
        filter_window: Window size of Savitzky-Golay filter. This should be odd number.
        filter_dim: Dimension of Savitzky-Golay filter.
        freq_guess: Optional. Initial frequency guess of :math:`F(x)`.

    Returns:
         Decay rate of signal.
    """
    y_smoothed = signal.savgol_filter(y, window_length=filter_window, polyorder=filter_dim)

    if freq_guess is not None and np.abs(freq_guess) > 0:
        period = 1 / np.abs(freq_guess)
        dt = np.mean(np.diff(x))
        width_samples = int(np.round(0.8 * period / dt))
    else:
        width_samples = 1

    peak_pos, _ = signal.find_peaks(y_smoothed, distance=width_samples)

    if len(peak_pos) < 2:
        return 0.0

    x_peaks = x[peak_pos]
    y_peaks = y_smoothed[peak_pos]

    return exp_decay(x_peaks, y_peaks)


def full_width_half_max(
    x: np.ndarray,
    y: np.ndarray,
    peak_index: int,
) -> float:
    """Get full width half maximum value of the peak. Offset of y should be removed.

    Args:
        x: Array of x values.
        y: Array of y values.
        peak_index: Index of peak.

    Returns:
        FWHM of the peak.

    Raises:
        AnalysisError: When peak is too broad and line width is not found.
    """
    y_ = np.abs(y)
    peak_height = y_[peak_index]
    halfmax_removed = np.sign(y_ - 0.5 * peak_height)

    try:
        r_bound = np.min(x[(halfmax_removed == -1) & (x > x[peak_index])])
    except ValueError:
        r_bound = None
    try:
        l_bound = np.max(x[(halfmax_removed == -1) & (x < x[peak_index])])
    except ValueError:
        l_bound = None

    if r_bound and l_bound:
        return r_bound - l_bound
    elif r_bound:
        return 2 * (r_bound - x[peak_index])
    elif l_bound:
        return 2 * (x[peak_index] - l_bound)

    raise AnalysisError("FWHM of input curve was not found. Perhaps scanning range is too narrow.")


def constant_spectral_offset(
    y: np.ndarray, filter_window: int = 5, filter_dim: int = 2, ratio: float = 0.1
) -> float:
    """Get constant offset of spectral baseline.

    This function searches constant offset by finding a region where 1st and 2nd order
    differentiation are close to zero. A return value is an average y value of that region.
    To suppress the noise contribution to derivatives, this function also applies a
    Savitzky-Golay filter to y value.

    This method is more robust to offset error than just taking median or average of y values
    especially when a peak width is wider compared to the scan range.

    Args:
        y: Array of y values.
        filter_window: Window size of Savitzky-Golay filter. This should be odd number.
        filter_dim: Dimension of Savitzky-Golay filter.
        ratio: Threshold value to decide flat region. This value represent a ratio
            to the maximum derivative value.

    Returns:
        Offset value.
    """
    y_smoothed = signal.savgol_filter(y, window_length=filter_window, polyorder=filter_dim)

    ydiff1 = np.abs(np.diff(y_smoothed, 1, append=np.nan))
    ydiff2 = np.abs(np.diff(y_smoothed, 2, append=np.nan, prepend=np.nan))
    non_peaks = y_smoothed[
        (ydiff1 < ratio * np.nanmax(ydiff1)) & (ydiff2 < ratio * np.nanmax(ydiff2))
    ]

    if len(non_peaks) == 0:
        return float(np.median(y))

    return np.average(non_peaks)


def constant_sinusoidal_offset(y: np.ndarray) -> float:
    """Get constant offset of sinusoidal signal.

    This function finds 95 and 5 percentile y values and take an average of them.
    This method is robust to the dependency on sampling window, i.e.
    if we sample sinusoidal signal for 2/3 of its period, simple averaging may induce
    a drift towards positive or negative direction depending on the phase offset.

    Args:
        y: Array of y values.

    Returns:
        Offset value.
    """
    maxv, _ = max_height(y, percentile=95)
    minv, _ = min_height(y, percentile=5)

    return 0.5 * (maxv + minv)


def rb_decay(
    x: np.ndarray,
    y: np.ndarray,
    b: float = 0.5,
) -> float:
    r"""Get base of exponential decay function which is assumed to be close to 1.

    This assumes following model:

    .. math::

        y(x) = a \alpha^x + b.

    To estimate the base of decay function :math:`\alpha`, we consider

    .. math::

        y'(x) = y(x) - b = a \alpha^x,

    and thus,

    .. math::

        y'(x+dx) = a \alpha^x \alpha^dx.

    By considering the ratio of y values at :math:`x+dx` to :math:`x`,

    .. math::

        ry = \frac{a \alpha^x \alpha^dx}{a \alpha^x} = \alpha^dx.

    From this relationship, we can estimate :math:`\alpha` as

    .. math::

        \alpha = ry^\frac{1}{dx}.

    Args:
        x: Array of x values.
        y: Array of y values.
        b: Asymptote of decay function.

    Returns:
         Base of decay function.
    """
    valid_inds = y > b

    # Remove y values below b
    y = y[valid_inds]
    x = x[valid_inds]

    if len(x) == 0:
        return 0
    if len(x) == 1:
        # If number of element is 1, assume y(0) = 1.0 and directly compute alpha.
        a = 1.0 - b
        return ((y[0] - b) / a) ** (1 / x[0])

    ry = (y[1:] - b) / (y[:-1] - b)
    dx = np.diff(x)

    return np.average(ry ** (1 / dx))



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

r"""
=========================================================================
Calibration Management (:mod:`qiskit_experiments.calibration_management`)
=========================================================================

.. currentmodule:: qiskit_experiments.calibration_management

.. warning::
    The calibrations interface is still in active development. It may have
    breaking API changes without deprecation warnings in future releases until
    otherwise indicated.

Calibrating qubit setups is the task of finding the pulse shapes and parameter
values that maximize the fidelity of the resulting quantum operations. This
therefore requires experiments which are analyzed to extract parameter values.
Furthermore, the resulting parameter values and schedules must be managed. The
calibration management module in Qiskit experiments allows users to manage
the resulting schedules and parameter values from obtained when running
calibration experiments from the :mod:`~qiskit_experiments.library`.

Classes
=======

.. autosummary::
    :toctree: ../stubs/

    Calibrations
    ParameterValue
    FixedFrequencyTransmon
    EchoedCrossResonance
    BasisGateLibrary
    BaseCalibrationExperiment


Managing Calibration Data
=========================

Calibrations are managed by the :class:`.Calibrations` class. This class stores schedules
which are intended to be fully parameterized, including the index of the channels. This
class:

* supports having different schedules share parameters
* allows default schedules for qubits that can be overridden for specific qubits.

The following code illustrates how a user can create a parameterized schedule, add
values to the parameters and query a schedule.

.. code-block:: python

    dur = Parameter("dur")
    amp = Parameter("amp")
    sigma = Parameter("σ")

    with pulse.build(name="xp") as xp:
        pulse.play(Gaussian(dur, amp, sigma), DriveChannel(Parameter("ch0")))

    cals = Calibrations()
    cals.add_schedule(xp)

    # add duration and sigma parameter values for all qubits.
    cals.add_parameter_value(160, "dur", schedule="xp")
    cals.add_parameter_value(35.5, "σ", schedule="xp")

    # Add an amplitude for qubit 3.
    cals.add_parameter_value(0.2+0.05j, "amp", (3, ), "xp")

    # Retrieve an xp pulse with all parameters assigned
    cals.get_schedule("xp", (3, ))

    # Retrieve an xp pulse with unassigned amplitude
    cals.get_schedule("xp", (3, ), free_params=["amp"])

The Calibrations make a couple of assumptions which are discussed below.

Parametric channel naming convention
************************************

Parametrized channel indices must be named according to a predefined pattern to properly
identify the channels and control channels when assigning values to the parametric
channel indices. A channel must have a name that starts with ``ch`` followed by an integer.
For control channels, this integer can be followed by a sequence ``.integer``.
Optionally, the name can end with ``$integer`` to specify the index of a control channel
for the case when a set of qubits share multiple control channels. For example,
valid channel names include ``"ch0"``, ``"ch1"``, ``"ch0.1"``, ``"ch0$"``, ``"ch2$3"``,
and ``"ch1.0.3$2"``.
The ``.`` delimiter is used to specify the different qubits when looking for control
channels. The optional $ delimiter is used to specify which control channel to use
if several control channels work together on the same qubits. For example, if the
control channel configuration is ``{(3,2): [ControlChannel(3), ControlChannel(12)]}``
then given qubits ``(2, 3)`` the name ``"ch1.0$1"`` will resolve to ``ControlChannel(12)`` while
``"ch1.0$0"`` will resolve to ``ControlChannel(3)``. A channel can only have one parameter.

Parameter naming restriction
****************************

Each parameter must have a unique name within each schedule. For example, it is
acceptable to have a parameter named ``amp`` in the schedule ``xp`` and a different
parameter instance named ``amp`` in the schedule named ``xm``. It is not acceptable
to have two parameters named ``amp`` in the same schedule. The naming restriction
only applies to parameters used in the immediate scope of the schedule. Schedules
called by Call instructions have their own scope for Parameter names.

The code block below illustrates the creation of a template schedule for a echoed cross-
resonance gate.

.. code-block:: python

    amp_cr = Parameter("amp")
    amp = Parameter("amp")
    d0 = DriveChannel(Parameter("ch0"))
    c1 = ControlChannel(Parameter("ch0.1"))
    sigma = Parameter("σ")
    width = Parameter("w")
    dur_xp = Parameter("duration")
    dur_cr = Parameter("duration")

    with pulse.build(name="xp") as xp:
        pulse.play(Gaussian(dur_xp, amp, sigma), d0)

    with pulse.build(name="cr") as cr:
        with pulse.align_sequential():
                pulse.play(GaussianSquare(dur_cr, amp_cr, sigma, width), c1)
                pulse.call(xp)
                pulse.play(GaussianSquare(dur_cr, -amp_cr, sigma, width), c1)
                pulse.call(xp)

    cals = Calibrations()
    cals.add_schedule(xp)
    cals.add_schedule(cr)

Note that a registered template schedule can be retrieve by doing

.. code-block:: python

    xp = cals.get_template("xp")

which would return the default xp schedule block template for all qubits.
"""

from .calibrations import Calibrations
from .parameter_value import ParameterValue
from .base_calibration_experiment import BaseCalibrationExperiment
from .basis_gate_library import FixedFrequencyTransmon, BasisGateLibrary, EchoedCrossResonance



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021, 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""
Utilities for using the Clifford group in randomized benchmarking.
"""

import itertools
import os
from functools import lru_cache
from numbers import Integral
from typing import Optional, Union, Tuple, Sequence, Iterable

import numpy as np

from qiskit.circuit import CircuitInstruction, Qubit
from qiskit.circuit import Gate, Instruction
from qiskit.circuit import QuantumCircuit, QuantumRegister
from qiskit.circuit.library import SdgGate, HGate, SGate, XGate, YGate, ZGate
from qiskit.compiler import transpile
from qiskit.exceptions import QiskitError
from qiskit.quantum_info import Clifford
from qiskit.transpiler import CouplingMap, PassManager
from qiskit.transpiler.passes.synthesis.high_level_synthesis import HLSConfig, HighLevelSynthesis

DEFAULT_SYNTHESIS_METHOD = "rb_default"

_DATA_FOLDER = os.path.join(os.path.dirname(__file__), "data")

_CLIFFORD_COMPOSE_1Q = np.load(f"{_DATA_FOLDER}/clifford_compose_1q.npz")["table"]
_CLIFFORD_INVERSE_1Q = np.load(f"{_DATA_FOLDER}/clifford_inverse_1q.npz")["table"]
_CLIFFORD_INVERSE_2Q = np.load(f"{_DATA_FOLDER}/clifford_inverse_2q.npz")["table"]
_clifford_compose_2q_data = np.load(f"{_DATA_FOLDER}/clifford_compose_2q_dense_selected.npz")
_CLIFFORD_COMPOSE_2Q_DENSE = _clifford_compose_2q_data["table"]
# valid indices for the columns of the _CLIFFORD_COMPOSE_2Q_DENSE table
_valid_sparse_indices = _clifford_compose_2q_data["valid_sparse_indices"]
# map a clifford number to the index of _CLIFFORD_COMPOSE_2Q_DENSE
_clifford_num_to_dense_index = {idx: ii for ii, idx in enumerate(_valid_sparse_indices)}
_CLIFFORD_TENSOR_1Q = np.load(f"{_DATA_FOLDER}/clifford_tensor_1q.npz")["table"]


# Transpilation utilities
def _transpile_clifford_circuit(
    circuit: QuantumCircuit, physical_qubits: Sequence[int]
) -> QuantumCircuit:
    # Simplified transpile that only decomposes Clifford circuits and creates the layout.
    return _apply_qubit_layout(_decompose_clifford_ops(circuit), physical_qubits=physical_qubits)


def _decompose_clifford_ops(circuit: QuantumCircuit) -> QuantumCircuit:
    # Simplified QuantumCircuit.decompose, which decomposes only Clifford ops
    res = circuit.copy_empty_like()
    if hasattr(circuit, "_parameter_table"):
        res._parameter_table = circuit._parameter_table
    for inst in circuit:
        if inst.operation.name.startswith("Clifford"):  # Decompose
            rule = inst.operation.definition.data
            if len(rule) == 1 and len(inst.qubits) == len(rule[0].qubits):
                if inst.operation.definition.global_phase:
                    res.global_phase += inst.operation.definition.global_phase
                res._data.append(
                    CircuitInstruction(
                        operation=rule[0].operation,
                        qubits=inst.qubits,
                        clbits=inst.clbits,
                    )
                )
            else:
                _circuit_compose(res, inst.operation.definition, qubits=inst.qubits)
        else:  # Keep the original instruction
            res._data.append(inst)
    return res


def _apply_qubit_layout(circuit: QuantumCircuit, physical_qubits: Sequence[int]) -> QuantumCircuit:
    # Mapping qubits in circuit to physical qubits (layout)
    res = QuantumCircuit(1 + max(physical_qubits), name=circuit.name, metadata=circuit.metadata)
    res.add_bits(circuit.clbits)
    for reg in circuit.cregs:
        res.add_register(reg)
    _circuit_compose(res, circuit, qubits=physical_qubits)
    if hasattr(circuit, "_parameter_table"):
        res._parameter_table = circuit._parameter_table
    return res


def _circuit_compose(
    self: QuantumCircuit, other: QuantumCircuit, qubits: Sequence[Union[Qubit, int]]
) -> QuantumCircuit:
    # Simplified QuantumCircuit.compose with clbits=None, front=False, inplace=True, wrap=False
    # without any validation, parameter_table/calibrations updates and copy of operations
    # The input circuit `self` is changed inplace.
    qubit_map = {
        other.qubits[i]: (self.qubits[q] if isinstance(q, int) else q) for i, q in enumerate(qubits)
    }
    for instr in other:
        self._data.append(
            CircuitInstruction(
                operation=instr.operation,
                qubits=[qubit_map[q] for q in instr.qubits],
                clbits=instr.clbits,
            ),
        )
    self.global_phase += other.global_phase
    return self


def _synthesize_clifford(
    clifford: Clifford,
    basis_gates: Optional[Tuple[str]],
    coupling_tuple: Optional[Tuple[Tuple[int, int]]] = None,
    synthesis_method: str = DEFAULT_SYNTHESIS_METHOD,
) -> QuantumCircuit:
    """Synthesize a circuit of a Clifford element. The resulting circuit contains only
    ``basis_gates`` and it complies with ``coupling_tuple``.

    Args:
        clifford: Clifford element to be converted
        basis_gates: basis gates to use in the conversion
        coupling_tuple: coupling map to use in the conversion in the form of tuple of edges
        synthesis_method: conversion algorithm name

    Returns:
        Synthesized circuit
    """
    qc = QuantumCircuit(clifford.num_qubits, name=str(clifford))
    qc.append(clifford, qc.qubits)
    return _synthesize_clifford_circuit(
        qc,
        basis_gates=basis_gates,
        coupling_tuple=coupling_tuple,
        synthesis_method=synthesis_method,
    )


def _synthesize_clifford_circuit(
    circuit: QuantumCircuit,
    basis_gates: Optional[Tuple[str]],
    coupling_tuple: Optional[Tuple[Tuple[int, int]]] = None,
    synthesis_method: str = DEFAULT_SYNTHESIS_METHOD,
) -> QuantumCircuit:
    """Convert a Clifford circuit into one composed of ``basis_gates`` with
    satisfying ``coupling_tuple`` using the specified synthesis method.

    Args:
        circuit: Clifford circuit to be converted
        basis_gates: basis gates to use in the conversion
        coupling_tuple: coupling map to use in the conversion in the form of tuple of edges
        synthesis_method: name of Clifford synthesis algorithm to use

    Returns:
        Synthesized circuit
    """
    if basis_gates:
        basis_gates = list(basis_gates)
    coupling_map = CouplingMap(coupling_tuple) if coupling_tuple else None

    # special handling for 1q or 2q case for speed
    if circuit.num_qubits <= 2:
        if synthesis_method == DEFAULT_SYNTHESIS_METHOD:
            return transpile(
                circuit,
                basis_gates=basis_gates,
                coupling_map=coupling_map,
                optimization_level=1,
            )
        else:
            # Provided custom synthesis method, re-synthesize Clifford circuit
            # convert the circuit back to a Clifford object and then call the synthesis plugin
            new_circuit = QuantumCircuit(circuit.num_qubits, name=circuit.name)
            new_circuit.append(Clifford(circuit), new_circuit.qubits)
            circuit = new_circuit

    # for 3q+ or custom synthesis method, synthesizes clifford circuit
    hls_config = HLSConfig(clifford=[(synthesis_method, {"basis_gates": basis_gates})])
    pm = PassManager([HighLevelSynthesis(hls_config=hls_config, coupling_map=coupling_map)])
    circuit = pm.run(circuit)
    return circuit


@lru_cache(maxsize=256)
def _clifford_1q_int_to_instruction(
    num: Integral,
    basis_gates: Optional[Tuple[str]],
    synthesis_method: str = DEFAULT_SYNTHESIS_METHOD,
) -> Instruction:
    return CliffordUtils.clifford_1_qubit_circuit(
        num, basis_gates=basis_gates, synthesis_method=synthesis_method
    ).to_instruction()


@lru_cache(maxsize=11520)
def _clifford_2q_int_to_instruction(
    num: Integral,
    basis_gates: Optional[Tuple[str]],
    coupling_tuple: Optional[Tuple[Tuple[int, int]]],
    synthesis_method: str = DEFAULT_SYNTHESIS_METHOD,
) -> Instruction:
    return CliffordUtils.clifford_2_qubit_circuit(
        num,
        basis_gates=basis_gates,
        coupling_tuple=coupling_tuple,
        synthesis_method=synthesis_method,
    ).to_instruction()


def _hash_cliff(cliff):
    return cliff.tableau.tobytes(), cliff.tableau.shape


def _dehash_cliff(cliff_hash):
    tableau = np.frombuffer(cliff_hash[0], dtype=bool).reshape(cliff_hash[1])
    return Clifford(tableau)


def _clifford_to_instruction(
    clifford: Clifford,
    basis_gates: Optional[Tuple[str]],
    coupling_tuple: Optional[Tuple[Tuple[int, int]]],
    synthesis_method: str = DEFAULT_SYNTHESIS_METHOD,
) -> Instruction:
    return _cached_clifford_to_instruction(
        _hash_cliff(clifford),
        basis_gates=basis_gates,
        coupling_tuple=coupling_tuple,
        synthesis_method=synthesis_method,
    )


@lru_cache(maxsize=256)
def _cached_clifford_to_instruction(
    cliff_hash: Tuple[str, Tuple[int, int]],
    basis_gates: Optional[Tuple[str]],
    coupling_tuple: Optional[Tuple[Tuple[int, int]]],
    synthesis_method: str = DEFAULT_SYNTHESIS_METHOD,
) -> Instruction:
    return _synthesize_clifford(
        _dehash_cliff(cliff_hash),
        basis_gates=basis_gates,
        coupling_tuple=coupling_tuple,
        synthesis_method=synthesis_method,
    ).to_instruction()


# The classes VGate and WGate are not actually used in the code - we leave them here to give
# a better understanding of the composition of the layers for 2-qubit Cliffords.
class VGate(Gate):
    """V Gate used in Clifford synthesis."""

    def __init__(self):
        """Create new V Gate."""
        super().__init__("v", 1, [])

    def _define(self):
        """V Gate definition."""
        q = QuantumRegister(1, "q")
        qc = QuantumCircuit(q)
        qc.data = [(SdgGate(), [q[0]], []), (HGate(), [q[0]], [])]
        self.definition = qc


class WGate(Gate):
    """W Gate used in Clifford synthesis."""

    def __init__(self):
        """Create new W Gate."""
        super().__init__("w", 1, [])

    def _define(self):
        """W Gate definition."""
        q = QuantumRegister(1, "q")
        qc = QuantumCircuit(q)
        qc.data = [(HGate(), [q[0]], []), (SGate(), [q[0]], [])]
        self.definition = qc


class CliffordUtils:
    """Utilities for generating one- and two-qubit Clifford circuits and elements."""

    NUM_CLIFFORD_1_QUBIT = 24
    NUM_CLIFFORD_2_QUBIT = 11520
    CLIFFORD_1_QUBIT_SIG = (2, 3, 4)
    CLIFFORD_2_QUBIT_SIGS = [  # TODO: deprecate
        (2, 2, 3, 3, 4, 4),
        (2, 2, 3, 3, 3, 3, 4, 4),
        (2, 2, 3, 3, 3, 3, 4, 4),
        (2, 2, 3, 3, 4, 4),
    ]

    @classmethod
    @lru_cache(maxsize=24)
    def clifford_1_qubit(cls, num):
        """Return the 1-qubit clifford element corresponding to `num`
        where `num` is between 0 and 23.
        """
        return Clifford(cls.clifford_1_qubit_circuit(num), validate=False)

    @classmethod
    @lru_cache(maxsize=11520)
    def clifford_2_qubit(cls, num):
        """Return the 2-qubit clifford element corresponding to ``num``,
        where ``num`` is between 0 and 11519.
        """
        return Clifford(cls.clifford_2_qubit_circuit(num), validate=False)

    @classmethod
    @lru_cache(maxsize=24)
    def clifford_1_qubit_circuit(
        cls,
        num,
        basis_gates: Optional[Tuple[str, ...]] = None,
        synthesis_method: str = DEFAULT_SYNTHESIS_METHOD,
    ):
        """Return the 1-qubit clifford circuit corresponding to ``num``,
        where ``num`` is between 0 and 23.
        """
        unpacked = cls._unpack_num(num, cls.CLIFFORD_1_QUBIT_SIG)
        i, j, p = unpacked[0], unpacked[1], unpacked[2]
        qc = QuantumCircuit(1, name=f"Clifford-1Q({num})")
        if i == 1:
            qc.h(0)
        if j == 1:
            qc.sxdg(0)
        if j == 2:
            qc.s(0)
        if p == 1:
            qc.x(0)
        if p == 2:
            qc.y(0)
        if p == 3:
            qc.z(0)

        if basis_gates:
            qc = _synthesize_clifford_circuit(qc, basis_gates, synthesis_method=synthesis_method)

        return qc

    @classmethod
    @lru_cache(maxsize=11520)
    def clifford_2_qubit_circuit(
        cls,
        num,
        basis_gates: Optional[Tuple[str, ...]] = None,
        coupling_tuple: Optional[Tuple[Tuple[int, int]]] = None,
        synthesis_method: str = DEFAULT_SYNTHESIS_METHOD,
    ):
        """Return the 2-qubit clifford circuit corresponding to `num`
        where `num` is between 0 and 11519.
        """
        qc = QuantumCircuit(2, name=f"Clifford-2Q({num})")
        for layer, idx in enumerate(_layer_indices_from_num(num)):
            if basis_gates:
                layer_circ = _transformed_clifford_layer(
                    layer, idx, basis_gates, coupling_tuple, synthesis_method=synthesis_method
                )
            else:
                layer_circ = _CLIFFORD_LAYER[layer][idx]
            _circuit_compose(qc, layer_circ, qubits=(0, 1))

        return qc

    @staticmethod
    def _unpack_num(num, sig):
        r"""Returns a tuple :math:`(a_1, \ldots, a_n)` where
        :math:`0 \le a_i \le \sigma_i` where
        sig=:math:`(\sigma_1, \ldots, \sigma_n)` and num is the sequential
        number of the tuple
        """
        res = []
        for k in sig:
            res.append(num % k)
            num //= k
        return res


# Constant mapping from 1Q single Clifford gate to 1Q Clifford numerical identifier.
# This table must be generated using `data.generate_clifford_data.gen_cliff_single_1q_gate_map`, or,
# equivalently, correspond to the ordering implicitly defined by CliffUtils.clifford_1_qubit_circuit.
_CLIFF_SINGLE_GATE_MAP_1Q = {
    ("id", (0,)): 0,
    ("h", (0,)): 1,
    ("sxdg", (0,)): 2,
    ("s", (0,)): 4,
    ("x", (0,)): 6,
    ("sx", (0,)): 8,
    ("y", (0,)): 12,
    ("z", (0,)): 18,
    ("sdg", (0,)): 22,
}
# Constant mapping from 2Q single Clifford gate to 2Q Clifford numerical identifier.
# This table must be generated using `data.generate_clifford_data.gen_cliff_single_2q_gate_map`, or,
# equivalently, correspond to the ordering defined by _layer_indices_from_num and _CLIFFORD_LAYER.
_CLIFF_SINGLE_GATE_MAP_2Q = {
    ("id", (0,)): 0,
    ("id", (1,)): 0,
    ("h", (0,)): 5760,
    ("h", (1,)): 2880,
    ("sxdg", (0,)): 6720,
    ("sxdg", (1,)): 3200,
    ("s", (0,)): 7680,
    ("s", (1,)): 3520,
    ("x", (0,)): 4,
    ("x", (1,)): 1,
    ("sx", (0,)): 6724,
    ("sx", (1,)): 3201,
    ("y", (0,)): 8,
    ("y", (1,)): 2,
    ("z", (0,)): 12,
    ("z", (1,)): 3,
    ("sdg", (0,)): 7692,
    ("sdg", (1,)): 3523,
    ("cx", (0, 1)): 16,
    ("cx", (1, 0)): 2336,
    ("cz", (0, 1)): 368,
    ("cz", (1, 0)): 368,
}


########
# Functions for 1-qubit integer Clifford operations
def compose_1q(lhs: Integral, rhs: Integral) -> Integral:
    """Return the composition of 1-qubit clifford integers."""
    return _CLIFFORD_COMPOSE_1Q[lhs, rhs]


def inverse_1q(num: Integral) -> Integral:
    """Return the inverse of a 1-qubit clifford integer."""
    return _CLIFFORD_INVERSE_1Q[num]


def num_from_1q_circuit(qc: QuantumCircuit) -> Integral:
    """Convert a given 1-qubit Clifford circuit to the corresponding integer.

    Note: The circuit must consist of gates in :const:`_CLIFF_SINGLE_GATE_MAP_1Q`,
    RZGate, Delay and Barrier.
    """
    num = 0
    for inst in qc:
        rhs = _num_from_1q_gate(op=inst.operation)
        num = _CLIFFORD_COMPOSE_1Q[num, rhs]
    return num


def _num_from_1q_gate(op: Instruction) -> int:
    """
    Convert a given 1-qubit clifford operation to the corresponding integer.
    Note that supported operations are limited to ones in :const:`_CLIFF_SINGLE_GATE_MAP_1Q` or Rz gate.

    Args:
        op: operation to be converted.

    Returns:
        An integer representing a Clifford consisting of a single operation.

    Raises:
        QiskitError: If the input instruction is not a Clifford instruction.
        QiskitError: If rz is given with a angle that is not Clifford.
    """
    if op.name in {"delay", "barrier"}:
        return 0
    try:
        name = _deparameterized_name(op)
        return _CLIFF_SINGLE_GATE_MAP_1Q[(name, (0,))]
    except QiskitError as err:
        raise QiskitError(
            f"Parameterized instruction {op.name} could not be converted to integer Clifford"
        ) from err
    except KeyError as err:
        raise QiskitError(
            f"Instruction {op.name} could not be converted to integer Clifford"
        ) from err


def _deparameterized_name(inst: Instruction) -> str:
    if inst.name == "rz":
        if np.isclose(inst.params[0], np.pi) or np.isclose(inst.params[0], -np.pi):
            return "z"
        elif np.isclose(inst.params[0], np.pi / 2):
            return "s"
        elif np.isclose(inst.params[0], -np.pi / 2):
            return "sdg"
        else:
            raise QiskitError(f"Wrong param {inst.params[0]} for rz in clifford")

    return inst.name


########
# Functions for 2-qubit integer Clifford operations
def compose_2q(lhs: Integral, rhs: Integral) -> Integral:
    """Return the composition of 2-qubit clifford integers."""
    num = lhs
    for layer, idx in enumerate(_layer_indices_from_num(rhs)):
        gate_numbers = _CLIFFORD_LAYER_NUMS[layer][idx]
        for n in gate_numbers:
            num = _CLIFFORD_COMPOSE_2Q_DENSE[num, _clifford_num_to_dense_index[n]]
    return num


def inverse_2q(num: Integral) -> Integral:
    """Return the inverse of a 2-qubit clifford integer."""
    return _CLIFFORD_INVERSE_2Q[num]


def num_from_2q_circuit(qc: QuantumCircuit) -> Integral:
    """Convert a given 2-qubit Clifford circuit to the corresponding integer.

    Note: The circuit must consist of gates in :const:`_CLIFF_SINGLE_GATE_MAP_2Q`,
    RZGate, Delay and Barrier.
    """
    lhs = 0
    for rhs in _clifford_2q_nums_from_2q_circuit(qc):
        lhs = _CLIFFORD_COMPOSE_2Q_DENSE[lhs, _clifford_num_to_dense_index[rhs]]
    return lhs


def _num_from_2q_gate(
    op: Instruction, qubits: Optional[Union[Tuple[int, int], Tuple[int]]] = None
) -> int:
    """
    Convert a given 1-qubit clifford operation to the corresponding integer.
    Note that supported operations are limited to ones in `_CLIFF_SINGLE_GATE_MAP_2Q` or Rz gate.

    Args:
        op: operation of instruction to be converted.
        qubits: qubits to which the operation applies

    Returns:
        An integer representing a Clifford consisting of a single operation.

    Raises:
        QiskitError: If the input instruction is not a Clifford instruction.
        QiskitError: If rz is given with a angle that is not Clifford.
    """
    if op.name in {"delay", "barrier"}:
        return 0

    qubits = qubits or (0, 1)
    try:
        name = _deparameterized_name(op)
        return _CLIFF_SINGLE_GATE_MAP_2Q[(name, qubits)]
    except QiskitError as err:
        raise QiskitError(
            f"Parameterized instruction {op.name} could not be converted to integer Clifford"
        ) from err
    except KeyError as err:
        raise QiskitError(
            f"Instruction {op.name} on {qubits} could not be converted to integer Clifford"
        ) from err


def _append_v_w(qc, vw0, vw1):
    if vw0 == "v":
        qc.sdg(0)
        qc.h(0)
    elif vw0 == "w":
        qc.h(0)
        qc.s(0)
    if vw1 == "v":
        qc.sdg(1)
        qc.h(1)
    elif vw1 == "w":
        qc.h(1)
        qc.s(1)


def _create_cliff_2q_layer_0():
    """Layer 0 consists of 0 or 1 H gates on each qubit, followed by 0/1/2 V gates on each qubit.
    Number of Cliffords == 36."""
    circuits = []
    num_h = [0, 1]
    v_w_gates = ["i", "v", "w"]
    for h0, h1, v0, v1 in itertools.product(num_h, num_h, v_w_gates, v_w_gates):
        qc = QuantumCircuit(2)
        for _ in range(h0):
            qc.h(0)
        for _ in range(h1):
            qc.h(1)
        _append_v_w(qc, v0, v1)
        circuits.append(qc)
    return circuits


def _create_cliff_2q_layer_1():
    """Layer 1 consists of one of the following:
    - nothing
    - cx(0,1) followed by 0/1/2 V gates on each qubit
    - cx(0,1), cx(1,0) followed by 0/1/2 V gates on each qubit
    - cx(0,1), cx(1,0), cx(0,1)
    Number of Cliffords == 20."""
    circuits = [QuantumCircuit(2)]  # identity at the beginning

    v_w_gates = ["i", "v", "w"]
    for v0, v1 in itertools.product(v_w_gates, v_w_gates):
        qc = QuantumCircuit(2)
        qc.cx(0, 1)
        _append_v_w(qc, v0, v1)
        circuits.append(qc)

    for v0, v1 in itertools.product(v_w_gates, v_w_gates):
        qc = QuantumCircuit(2)
        qc.cx(0, 1)
        qc.cx(1, 0)
        _append_v_w(qc, v0, v1)
        circuits.append(qc)

    qc = QuantumCircuit(2)  # swap at the end
    qc.cx(0, 1)
    qc.cx(1, 0)
    qc.cx(0, 1)
    circuits.append(qc)
    return circuits


def _create_cliff_2q_layer_2():
    """Layer 2 consists of a Pauli gate on each qubit {Id, X, Y, Z}.
    Number of Cliffords == 16."""
    circuits = []
    pauli = ("i", XGate(), YGate(), ZGate())
    for p0, p1 in itertools.product(pauli, pauli):
        qc = QuantumCircuit(2)
        if p0 != "i":
            qc.append(p0, [0])
        if p1 != "i":
            qc.append(p1, [1])
        circuits.append(qc)
    return circuits


_CLIFFORD_LAYER = (
    _create_cliff_2q_layer_0(),
    _create_cliff_2q_layer_1(),
    _create_cliff_2q_layer_2(),
)
_NUM_LAYER_1 = 20
_NUM_LAYER_2 = 16


def _clifford_2q_nums_from_2q_circuit(qc: QuantumCircuit) -> Iterable[Integral]:
    """Yield Clifford numbers that represents the 2Q Clifford circuit."""
    for inst in qc:
        qubits = tuple(qc.find_bit(q).index for q in inst.qubits)
        yield _num_from_2q_gate(op=inst.operation, qubits=qubits)


# Construct mapping from Clifford layers to series of Clifford numbers
_CLIFFORD_LAYER_NUMS = [
    [tuple(_clifford_2q_nums_from_2q_circuit(qc)) for qc in _CLIFFORD_LAYER[layer]]
    for layer in [0, 1, 2]
]


@lru_cache(maxsize=256)
def _transformed_clifford_layer(
    layer: int,
    index: Integral,
    basis_gates: Tuple[str, ...],
    coupling_tuple: Optional[Tuple[Tuple[int, int]]],
    synthesis_method: str = DEFAULT_SYNTHESIS_METHOD,
) -> QuantumCircuit:
    # Return the index-th quantum circuit of the layer translated with the basis_gates.
    # The result is cached for speed.
    return _synthesize_clifford_circuit(
        _CLIFFORD_LAYER[layer][index],
        basis_gates=basis_gates,
        coupling_tuple=coupling_tuple,
        synthesis_method=synthesis_method,
    )


def _num_from_layer_indices(triplet: Tuple[Integral, Integral, Integral]) -> Integral:
    """Return the clifford number corresponding to the input triplet."""
    num = triplet[0] * _NUM_LAYER_1 * _NUM_LAYER_2 + triplet[1] * _NUM_LAYER_2 + triplet[2]
    return num


def _layer_indices_from_num(num: Integral) -> Tuple[Integral, Integral, Integral]:
    """Return the triplet of layer indices corresponding to the input number."""
    idx2 = num % _NUM_LAYER_2
    num = num // _NUM_LAYER_2
    idx1 = num % _NUM_LAYER_1
    idx0 = num // _NUM_LAYER_1
    return idx0, idx1, idx2


def _tensor_1q_nums(first: Integral, second: Integral) -> Integral:
    """Return the 2-qubit Clifford integer that is the tensor product of 1-qubit Cliffords."""
    return _CLIFFORD_TENSOR_1Q[first, second]



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021, 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""
Correlated readout error calibration experiment class.
"""
from typing import Iterable, List, Optional
from qiskit import QuantumCircuit
from qiskit.providers.backend import BackendV2, Backend
from qiskit.exceptions import QiskitError
from qiskit_experiments.framework import BaseExperiment
from qiskit_experiments.library.characterization.analysis.correlated_readout_error_analysis import (
    CorrelatedReadoutErrorAnalysis,
)


def calibration_circuit(num_qubits: int, state_label: str) -> QuantumCircuit:
    """Return a calibration circuit.

    This is an N-qubit circuit where N is the length of the label.
    The circuit consists of X-gates on qubits with label bits equal to 1,
    and measurements of all qubits.
    """
    circ = QuantumCircuit(num_qubits, name="meas_mit_cal_" + state_label)
    for i, val in enumerate(reversed(state_label)):
        if val == "1":
            circ.x(i)
    circ.measure_all()
    circ.metadata = {"state_label": state_label}
    return circ


class CorrelatedReadoutError(BaseExperiment):
    r"""Correlated readout error characterization experiment.

    # section: overview
        This class constructs the a :class:`~qiskit.result.CorrelatedReadoutMitigator`
        containing the full assignment matrix :math:`A` characterizing the readout error
        for the given qubits from the experiment results accessible via the
        :meth:`~qiskit.result.CorrelatedReadoutMitigator.assignment_matrix` method.

        Readout errors affect quantum computation during the measurement
        of the qubits in a quantum device. By characterizing the readout errors,
        it is possible to construct a *readout error mitigator* that is used both
        to obtain a more accurate distribution of the outputs, and more accurate
        measurements of expectation value for measurables.

        The readout mitigator is generated from an *assignment matrix*:
        a :math:`2^n\times 2^n` matrix :math:`A` such that :math:`A_{y,x}` is the probability
        to observe :math:`y` given the true outcome should be :math:`x`. The assignment matrix is used
        to compute the *mitigation matrix* used in the readout error mitigation process itself.

        A *Correlated readout mitigator* uses the full :math:`2^n \times 2^n` assignment matrix, meaning
        it can only be used for small values of :math:`n`.
        The corresponding class in Qiskit is the
        :class:`~qiskit.result.CorrelatedReadoutMitigator` in :mod:`qiskit.result`.

        The experiment generates :math:`2^n` circuits, for every possible
        :math:`n`-qubit quantum state and constructs
        the assignment matrix and correlated mitigator from the results.

        See :class:`CorrelatedReadoutErrorAnalysis`
        documentation for additional information on correlated readout error experiment analysis.

    # section: analysis_ref
        :class:`CorrelatedReadoutErrorAnalysis`

    # section: example
        .. jupyter-execute::
            :hide-code:

            # Temporary workaround for missing support in Qiskit and qiskit-ibm-runtime
            from qiskit_experiments.test.patching import patch_sampler_test_support
            patch_sampler_test_support()

            from qiskit.providers.fake_provider import GenericBackendV2
            from qiskit_aer import AerSimulator

            num_qubits=5
            backend = AerSimulator.from_backend(GenericBackendV2(num_qubits=num_qubits,
                                                                 calibrate_instructions=True))

        .. jupyter-execute::

            from qiskit_experiments.library import CorrelatedReadoutError

            exp = CorrelatedReadoutError(physical_qubits=(0,1,2), backend=backend)

            exp_data = exp.run().block_for_results()
            display(exp_data.figure(0))
            exp_data.analysis_results(dataframe=True)

    # section: manual
        :doc:`/manuals/measurement/readout_mitigation`

    # section: reference
        .. ref_arxiv:: 1 2006.14044
    """

    def __init__(
        self,
        physical_qubits: Optional[Iterable[int]] = None,
        backend: Optional[Backend] = None,
    ):
        """Initialize a correlated readout error characterization experiment.

        Args:
            physical_qubits: Optional, the backend qubits being characterized
                for readout error. If None all qubits on the provided backend
                will be characterized.
            backend: Optional, the backend to characterize.

        Raises:
            QiskitError: If args are not valid.
        """
        if physical_qubits is None:
            if backend is None:
                raise QiskitError("`physical_qubits` and `backend` kwargs cannot both be None.")
            num_qubits = 0
            if isinstance(backend, BackendV2):
                num_qubits = backend.target.num_qubits
            elif isinstance(backend, Backend):
                num_qubits = backend.configuration().num_qubits
            if num_qubits:
                physical_qubits = range(num_qubits)
            else:
                raise QiskitError(f"Cannot infer backend qubits from backend {backend}")
        super().__init__(physical_qubits, backend=backend)
        self.analysis = CorrelatedReadoutErrorAnalysis()

    def circuits(self) -> List[QuantumCircuit]:
        """Returns the experiment's circuits"""
        labels = [bin(j)[2:].zfill(self.num_qubits) for j in range(2**self.num_qubits)]
        return [calibration_circuit(self.num_qubits, label) for label in labels]



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
======================================================
Experiment Library (:mod:`qiskit_experiments.library`)
======================================================

.. currentmodule:: qiskit_experiments.library

A library of quantum characterization, calibration and verification
experiments for calibrating and benchmarking quantum devices. See
:mod:`qiskit_experiments.framework` for general information on the framework
for running experiments.


.. _verification:

Verification Experiments
========================

Experiments for verification and validation of quantum devices.

.. autosummary::
    :toctree: ../stubs/
    :template: autosummary/experiment.rst

    ~randomized_benchmarking.StandardRB
    ~randomized_benchmarking.InterleavedRB
    ~tomography.TomographyExperiment
    ~tomography.StateTomography
    ~tomography.ProcessTomography
    ~tomography.MitigatedStateTomography
    ~tomography.MitigatedProcessTomography
    ~quantum_volume.QuantumVolume

.. _characterization single qubit:

Characterization Experiments: Single Qubit
==========================================

Experiments for characterization of properties of individual qubits.
Some experiments also have a calibration experiment version.

.. autosummary::
    :toctree: ../stubs/
    :template: autosummary/experiment.rst

    ~characterization.T1
    ~characterization.T2Hahn
    ~characterization.T2Ramsey
    ~characterization.Tphi
    ~characterization.QubitSpectroscopy
    ~characterization.EFSpectroscopy
    ~characterization.HalfAngle
    ~characterization.FineAmplitude
    ~characterization.FineXAmplitude
    ~characterization.FineSXAmplitude
    ~characterization.Rabi
    ~characterization.EFRabi
    ~characterization.RamseyXY
    ~characterization.FineFrequency
    ~characterization.ReadoutAngle
    ~characterization.ResonatorSpectroscopy
    ~characterization.RoughDrag
    ~characterization.FineDrag
    ~characterization.FineXDrag
    ~characterization.FineSXDrag
    ~characterization.MultiStateDiscrimination
    ~driven_freq_tuning.StarkRamseyXY
    ~driven_freq_tuning.StarkRamseyXYAmpScan
    ~driven_freq_tuning.StarkP1Spectroscopy

.. _characterization two qubits:

Characterization Experiments: Two Qubits
========================================

Experiments for characterization of properties of two qubit interactions.

.. autosummary::
    :toctree: ../stubs/
    :template: autosummary/experiment.rst

    ~characterization.CrossResonanceHamiltonian
    ~characterization.EchoedCrossResonanceHamiltonian
    ~characterization.ZZRamsey
    ~characterization.FineZXAmplitude

.. _characterization-mitigation:

Characterization Experiments: Mitigation
========================================

Experiments for characterizing and mitigating readout error.

.. autosummary::
    :toctree: ../stubs/
    :template: autosummary/experiment.rst

    ~characterization.LocalReadoutError
    ~characterization.CorrelatedReadoutError

.. _calibration:

Calibration Experiments
=======================

Experiments for pulse level calibration of quantum gates. These experiments
are usually run with a
:class:`~qiskit_experiments.calibration_management.Calibrations`
class instance to manage parameters and pulse schedules.
See :doc:`/tutorials/calibrations` for examples.

.. autosummary::
    :toctree: ../stubs/
    :template: autosummary/experiment.rst

    ~calibration.RoughFrequencyCal
    ~calibration.RoughEFFrequencyCal
    ~calibration.FrequencyCal
    ~calibration.FineFrequencyCal
    ~calibration.RoughDragCal
    ~calibration.FineXDragCal
    ~calibration.FineSXDragCal
    ~calibration.FineDragCal
    ~calibration.FineAmplitudeCal
    ~calibration.FineXAmplitudeCal
    ~calibration.FineSXAmplitudeCal
    ~calibration.HalfAngleCal
    ~calibration.RoughAmplitudeCal
    ~calibration.RoughXSXAmplitudeCal
    ~calibration.EFRoughXSXAmplitudeCal

"""
from .calibration import (
    RoughDragCal,
    FineDragCal,
    FineXDragCal,
    FineSXDragCal,
    RoughAmplitudeCal,
    RoughXSXAmplitudeCal,
    EFRoughXSXAmplitudeCal,
    FineAmplitudeCal,
    FineXAmplitudeCal,
    FineSXAmplitudeCal,
    RoughFrequencyCal,
    RoughEFFrequencyCal,
    FrequencyCal,
    FineFrequencyCal,
    HalfAngleCal,
)
from .characterization import (
    T1,
    T2Hahn,
    T2Ramsey,
    Tphi,
    QubitSpectroscopy,
    EFSpectroscopy,
    CrossResonanceHamiltonian,
    EchoedCrossResonanceHamiltonian,
    RoughDrag,
    FineDrag,
    FineXDrag,
    FineSXDrag,
    Rabi,
    EFRabi,
    HalfAngle,
    FineAmplitude,
    FineXAmplitude,
    FineSXAmplitude,
    FineZXAmplitude,
    RamseyXY,
    FineFrequency,
    ReadoutAngle,
    ResonatorSpectroscopy,
    LocalReadoutError,
    CorrelatedReadoutError,
    ZZRamsey,
    MultiStateDiscrimination,
)
from .randomized_benchmarking import StandardRB, InterleavedRB
from .tomography import (
    TomographyExperiment,
    StateTomography,
    ProcessTomography,
    MitigatedStateTomography,
    MitigatedProcessTomography,
)
from .quantum_volume import QuantumVolume
from .driven_freq_tuning import (
    StarkRamseyXY,
    StarkRamseyXYAmpScan,
    StarkP1Spectroscopy,
)

# Experiment Sub-modules
from . import calibration
from . import characterization
from . import randomized_benchmarking
from . import tomography
from . import quantum_volume
